#include <Fonts.h>
#include <Menus.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Memory.h>
#include <Packages.h>
#include <Errors.h>

#ifndef __OCEAUTHDIR__
#include <OCEAuthDir.h>
#endif

#ifndef __OCEERRORS__
#include <OCEErrors.h>
#endif

#ifndef __FILES__
#include <Files.h>
#endif



#define kEnumBufferSize 1024
#define kInitialLookupBuffer 256

struct LookupInfo {
	unsigned long		fileID;
	Boolean				found;
};

#ifndef __cplusplus
typedef struct LookupInfo LookupInfo;
#endif

struct EnumInfo {
	short				setupRef;
	unsigned long		fileID;
	CreationID			msamCID;
	DirEnumSpec			lastEnumSpec;
	RString				name;
	RString				type;
};

#ifndef __cplusplus
typedef struct EnumInfo EnumInfo;
#endif




// Return the given file's file ID
OSErr GetIDFromFSSpec(const FSSpec *spec, unsigned long *id)
	{
	OSErr err;
	CInfoPBRec pb;
	
	pb.dirInfo.ioCompletion = nil;
	pb.dirInfo.ioVRefNum = spec->vRefNum;
	pb.dirInfo.ioNamePtr = spec->name;
	pb.dirInfo.ioFDirIndex = 0;
	pb.dirInfo.ioDrDirID = spec->parID;
	err = PBGetCatInfoSync(&pb);
	if (err == noErr)
		{
		*id = pb.dirInfo.ioDrDirID;
		}
	return err;
	}



// Return the dsRefNum of the Setup Directory
OSErr GetSetupDirectoryRefNum(short *refNum)
	{
	OSErr err;
	DirParamBlock dspb;
	
	err = DirGetOCESetupRefNum(&dspb, false);
	
	if (err == noErr)
		{
		*refNum = dspb.header.dsRefNum;
		}
	
	return err;
	}



// Does this record have our file ID as an attribute?
pascal Boolean LookupCB(long info, const Attribute *thisAttribute)
LookupInfo* info;
	{
	if (info->fileID == * (long*) thisAttribute->value.bytes)
		{
		info->found = true;
		}

	// Stop the parse once we find an attribute value with our file ID.
	return info->found;
	}

// Do a lookup into the given record in the Setup Directory to
// see if its kGatewayFileIDAttrTypeNum attribute stores the given
// file ID.
Boolean RecordHasFileID(const LocalRecordID *lrid, short setupRef, unsigned long fileID)
	{
	OSErr err;
	LookupInfo info;
	DirParamBlock dspb;
	RecordID rid;
	RecordIDPtr recordList[1];
	AttributeTypePtr attrTypeList[1];
	Boolean includeStartingPoint = false;
	unsigned long bufferSize = kInitialLookupBuffer;
	void *dataBuffer;
	

	info.fileID = fileID;
	info.found = false;
	
	rid.local = *lrid;		// The lookup requires a RID, not a local RID.
	rid.rli = nil;
	
	recordList[0] = &rid;
	attrTypeList[0] = OCEGetIndAttributeType(kGatewayFileIDAttrTypeNum);

	// Create a buffer that's big enough to get at least one attribute value
	dataBuffer = NewPtr(bufferSize);
	err = MemError();
	if (err == noErr)
		{
		*(long *)&dspb.lookupGetPB.serverHint = 0;
		dspb.lookupGetPB.dsRefNum = setupRef;
		dspb.lookupGetPB.aRecordList = recordList;
		dspb.lookupGetPB.attrTypeList = attrTypeList;
		dspb.lookupGetPB.recordIDCount = 1;
		dspb.lookupGetPB.attrTypeCount = 1;
		dspb.lookupGetPB.includeStartingPoint = false;
		dspb.lookupGetPB.getBuffer = dataBuffer;
		dspb.lookupGetPB.getBufferSize = bufferSize;
		dspb.lookupGetPB.startingRecordIndex = 1;
		dspb.lookupGetPB.startingAttrTypeIndex = 1;
		OCESetCreationIDtoNull(&dspb.lookupGetPB.startingAttribute.cid);
	
		dspb.lookupParsePB.clientData = &info;
		dspb.lookupParsePB.eachRecordID = nil;
		dspb.lookupParsePB.eachAttrType = nil;
		dspb.lookupParsePB.eachAttrValue = LookupCB;
		
		do
			{
			err = DirLookupGet(&dspb,false);
			
			if ((err == noErr) || (err == kOCEMoreData))
				{
				err = DirLookupParse(&dspb,false);
				}
			} while (err == kOCEMoreData);

		DisposePtr((Ptr) dataBuffer);
		}

	return info.found;
	}



// Check to see if the given record is our MSAM record. We tell
// this by checking to see if its kGatewayFileIDAttrTypeNum attribute
// is the file ID of our MSAM file.
pascal Boolean EnumCB(long enumData, const DirEnumSpec *enumSpecPtr)
EnumInfo *enumData;
	{
	Boolean found;

	// First save our current DirEnumSpec so that we can continue
	// if our buffer couldn't hold all the records.
	BlockMove(enumSpecPtr, &enumData->lastEnumSpec, sizeof(DirEnumSpec));
	OCECopyRString(enumSpecPtr->u.recordIdentifier.recordName, &enumData->name, kRStringMaxBytes);
	OCECopyRString(enumSpecPtr->u.recordIdentifier.recordType, &enumData->type, kRStringMaxBytes);
	enumData->lastEnumSpec.u.recordIdentifier.recordName = &enumData->name;
	enumData->lastEnumSpec.u.recordIdentifier.recordType = &enumData->type;
	
	// Does this record have our file ID?
	found = RecordHasFileID(&enumSpecPtr->u.recordIdentifier, enumData->setupRef, enumData->fileID);
	
	// If so, save away it's CID.
	if (found)
		{
		OCECopyCreationID(&enumSpecPtr->u.recordIdentifier.cid, &enumData->msamCID);
		}
	
	// Stop the parse once we find a record with our file ID.
	return found;
	}


// Enumerate all MSAM records in the Setup Directory, looking
// for one that corresponds to the MSAM with the given file ID.
OSErr FindMSAMRecordWithFileID(short setupRef, unsigned long fileID, CreationID *msamCID)
	{
	OSErr err;
	DirParamBlock dspb;
	void *buffer;
	RString *recordType;
	EnumInfo enumData;
	
	
	buffer = NewPtr(kEnumBufferSize);
	err = MemError();

	if (err == noErr)
		{
		recordType = OCEGetIndRecordType(kMSAMRecTypeNum);
		
		enumData.fileID = fileID;
		enumData.setupRef = setupRef;
		OCESetCreationIDtoNull(&enumData.msamCID);

		*(long *)&dspb.enumerateGetPB.serverHint = 0;
		dspb.enumerateGetPB.dsRefNum = setupRef;
		dspb.enumerateGetPB.clientData = &enumData;
		dspb.enumerateGetPB.aRLI = nil;
		dspb.enumerateGetPB.startingPoint = nil;
		dspb.enumerateGetPB.sortBy = kSortByName;
		dspb.enumerateGetPB.sortDirection = kSortForwards;
		dspb.enumerateGetPB.nameMatchString = nil;
		dspb.enumerateGetPB.typesList = &recordType;
		dspb.enumerateGetPB.typeCount = 1;
		dspb.enumerateGetPB.enumFlags = kEnumDistinguishedNameMask;
		dspb.enumerateGetPB.includeStartingPoint = false;
		dspb.enumerateGetPB.matchNameHow = kMatchAll;
		dspb.enumerateGetPB.matchTypeHow = kExactMatch;
		dspb.enumerateGetPB.getBuffer = buffer;
		dspb.enumerateGetPB.getBufferSize = kEnumBufferSize;

		dspb.enumerateParsePB.eachEnumSpec = EnumCB;
	
		do
			{
			err = DirEnumerateGet(&dspb, false);
			if ((err == noErr) || (err == kOCEMoreData))
				{
				err = DirEnumerateParse(&dspb, false);
				}
			dspb.enumerateGetPB.startingPoint = &enumData.lastEnumSpec;
			} while (err == kOCEMoreData);
			
		DisposPtr((Ptr) buffer);
		}
	
	OCECopyCreationID(&enumData.msamCID, msamCID);
	
	return err;
	}



// Create an MSAM record, returning the record's CID.
OSErr CreateMSAMRecord(short setupRef, CreationID *msamCID)
	{
	OSErr err;
	RString name;
	RecordID rid;
	DirParamBlock dspb;
	
	OCEPToRString("\pMy MSAM", smRoman, &name, kRStringMaxBytes);
	rid.local.recordName = &name;
	rid.local.recordType = OCEGetIndRecordType(kMSAMRecTypeNum);
	OCESetCreationIDtoNull(&rid.local.cid);
	rid.rli = nil;

	*(long *)&dspb.addRecordPB.serverHint = 0;
	dspb.addRecordPB.dsRefNum = setupRef;
	dspb.addRecordPB.aRecord = &rid;
	dspb.addRecordPB.allowDuplicate = true;
	
	err = DirAddRecord(&dspb, false);

	OCECopyCreationID(&rid.local.cid, msamCID);
	
	return err;
	}
	
	

// Add an attribute value to the given record in the Setup Directory.
OSErr AddAttribute(short setupRef, 
				const CreationID *recordCID, 
				const AttributeType *attrType, 
				unsigned long length, 
				Ptr bytes)
	{
	OSErr err;
	RecordID rid;
	Attribute attr;
	DirParamBlock dspb;
	
	rid.local.recordName = nil;
	rid.local.recordType = nil;
	OCECopyCreationID(recordCID, &rid.local.cid);
	rid.rli = nil;
	
	OCECopyRString((RString*) attrType, (RString*) &attr.attributeType, kAttributeTypeMaxBytes);
	OCESetCreationIDtoNull(&attr.cid);
	attr.value.tag = typeBinary;
	attr.value.dataLength = length;
	attr.value.bytes = bytes;
	
	*(long *)&dspb.addAttributeValuePB.serverHint = 0;
	dspb.addAttributeValuePB.dsRefNum = setupRef;
	dspb.addAttributeValuePB.aRecord = &rid;
	dspb.addAttributeValuePB.attr = &attr;
	
	err = DirAddAttributeValue(&dspb, false);

	return err;
	}

// Given an FSSpec representing an MSAM file, return the corresponding
// MSAM record's CID, creating the MSAM record if necessary.
OSErr GetMSAMCreationID(const FSSpec *spec, CreationID *msamCID)
	{
	OSErr err;
	unsigned fileID;
	short setupRef;
	long version = 1;
	
	err = GetIDFromFSSpec(spec, &fileID);
	
	if (err == noErr)
		{
		err = GetSetupDirectoryRefNum(&setupRef);
		}
	
	if (err == noErr)
		{
		err = FindMSAMRecordWithFileID(setupRef, fileID, msamCID);
		}
	
	// If we couldn't find the record, create it
	if ((err == noErr) && OCEEqualCreationID(msamCID, OCENullCID()))
		{
		err = CreateMSAMRecord(setupRef, msamCID);
		
		if (err == noErr)
			{
			// Add the file ID attribute
			err = AddAttribute(setupRef, msamCID, OCEGetIndAttributeType(kGatewayFileIDAttrTypeNum), sizeof(long), (Ptr) &fileID);
			}

		if (err == noErr)
			{
			// Add the version attribute.
			err = AddAttribute(setupRef, msamCID, OCEGetIndAttributeType(kVersionAttrTypeNum), sizeof(long), (Ptr) &version);
			}
		}

	return err;
	}









OSErr AddRecordReference(const CreationID *recordToAddReference,
						 const AttributeType *attrType,
						 const CreationID *referenceCID)
	{
	OSErr err;
	short setupRef;
	RecordID recordReference;
	PackedRecordID *packedReference;
	unsigned short size;
	
	err = GetSetupDirectoryRefNum(&setupRef);
	
	if (err == noErr)
		{
		recordReference.local.recordName = nil;
		recordReference.local.recordType = nil;
		OCECopyCreationID(referenceCID, &recordReference.local.cid);
		recordReference.rli = nil;
		
		size = OCEPackedRecordIDSize(&recordReference);
		packedReference = (PackedRecordID*) NewPtr(size);
		err = MemError();
		}
	
	if (err == noErr)
		{
		OCEPackRecordID(&recordReference, packedReference, size);
		err = AddAttribute(setupRef, recordToAddReference, attrType, size, (Ptr) packedReference);
		}

	return err;
	}










void Initialize()
{
	InitGraf((Ptr) &qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
} /*Initialize*/

main()
	{
	OSErr err;
	StandardFileReply reply;
	CreationID cid;
	
	Debugger();
	MaxApplZone();					/* expand the heap so code segments load at the top */

	Initialize();					/* initialize the program */
	
	StandardGetFile(nil, 0, nil, &reply);
	if (reply.sfGood)
		{
		err = GetMSAMCreationID(&reply.sfFile, &cid);
		Debugger();
		}
	}

