/*							AttributeTypeListManager.c							*/
/*
 * DTS Catalog Peek
 * AttributeTypeListManager.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions manage the attribute type list.
 *
 * Attribute Type List Globals:
 *	DBR.attributeTypeList
 *			ListManager list that stores attribute types.
 *			Each entry is an AttributeType (i.e., RString).
 *	DBR.currentAttributeTypeIndex
 *			The attribute type that the user has selected.
 *			Initialized to zero (first attribute type) when'
 *			the list is initialized, or -1 if the list is empty.
 *	DBR.currentAttrType
 *			A copy of the currently-selected attribute type.
 *			This is the attribute type that will be examined
 *			by EnumerateAttributeValues
 * DBR.startingAttrType
 *			A copy of the last attribute type in the list.
 *			This is needed by the enumeration. If there is
 *			no currently-selected attribute type, it
 *			will be set to a null-string.
 */
#include "DTSCatalogPeek.h"
#pragma segment MainCode

#define TYPE_LIST	(DBR.attributeTypeList)
#define TYPE_BOUNDS	((**TYPE_LIST).dataBounds)

/*
 * Local function prototypes
 */
Boolean								GetNthAttributeType(
		register CatalogBrowserPtr		dbp,
		short							cellIndex,
		Boolean							setSelect,
		AttributeTypePtr				result
	);

/*
 * ClearAttributeTypeList
 * Clear the attribute type list and, en-passant, the
 * attribute value list. This is called when the user
 * selects a new directory entry.
 */
void
ClearAttributeTypeList(
		register CatalogBrowserPtr		dbp
	)
{
		FailNIL(DBR.attributeTypeList);
		LDelRow(0, 0, DBR.attributeTypeList);
		DBR.currentAttributeTypeIndex = -1;
		ClearAttributeValueList(dbp);
}

/*
 * DoClickInAttributeTypeList
 * The user clicked in the attribute type list.
 * Get the new selection, ignoring double-clicks.
 * If the selection changed, enumerate the attribute values.
 */
void
DoClickInAttributeTypeList(
		register CatalogBrowserPtr		dbp
	)
{
		Cell				theCell;

		LClick(DBR.penPt, EVENT.modifiers, TYPE_LIST);
		theCell = LLastClick(TYPE_LIST);
		if (PtInRect(theCell, &TYPE_BOUNDS)
		 && theCell.v != DBR.currentAttributeTypeIndex) {
			ClearAttributeValueList(dbp);
			SelectNthAttributeType(dbp, theCell.v);
		}
}

/*
 * SelectNthAttributeType
 * Select a specific attribute type and enumerate its values.
 */
void
SelectNthAttributeType(
		register CatalogBrowserPtr		dbp,
		short							index
	)
{
		if (GetNthAttributeType(dbp, index, TRUE, &DBR.currentAttrType)) {
			DBR.currentAttributeTypeIndex = index;
			EnumerateAttributeValues(dbp);
		}
}

/*
 * AddAttributeType
 * This is called by the attribute type enumeration parse
 * callback. It stores the new attribute at the end of the
 * list. A "real" program would sort the values or something.
 * Note, however, that the last attribute type stored is
 * needed to restart the Get/Parse enumeration.
 *
 * We add the data at DBR.currentAttributeType
 */
void
AddAttributeType(
		register CatalogBrowserPtr		dbp
	)
{
		Cell				theCell;
		
		theCell.v = LAddRow(1, TYPE_BOUNDS.bottom, TYPE_LIST);
		FailMemError();
		theCell.h = 0;
		LSetCell(
			(Ptr) &DBR.currentAttrType,
			RStringLength(&DBR.currentAttrType),
			theCell,
			TYPE_LIST
		);
		FailMemError();
}

/*
 * GetNthAttributeType
 * Get selected attribute type, return TRUE if there was one.
 */
Boolean
GetNthAttributeType(
		register CatalogBrowserPtr		dbp,
		short							cellIndex,
		Boolean							setSelect,
		AttributeTypePtr				result
	)
{
		static const AttributeType		nullAttribute;		/* Null datum	*/
		short							dataLen;
		Cell							theCell;
		Boolean							gotOne;
		
		SetPt(&theCell, 0, cellIndex);
		if (PtInRect(theCell, &TYPE_BOUNDS) == FALSE) {
		 	*result = nullAttribute;
		 	gotOne = FALSE;
		}
		else {
			dataLen = sizeof (AttributeType);
			LGetCell(result, &dataLen, theCell, TYPE_LIST);
			if (setSelect)
				LSetSelect(TRUE, theCell, TYPE_LIST);
			gotOne = TRUE;
		}
		return (gotOne);
}
		
