/*							AttributeValueListManager.c							*/
/*
 * DTS Catalog Peek
 * AttributeValueListManager.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions manage the attribute value list.
 *
 * Attribute Value List Globals:
 *	DBR.attributeValueList
 *			ListManager list that stores attribute values.
 *			Each entry is a structure that contains the
 *			AttributeType (RString), creation id (bytes),
 *			tag (OSType), and a handle containing the
 *			actual data.
 * DBR.currentAttributeValueIndex
 *			The index of the currently-selected attribute
 *			value, or -1 if no value is selected or the
 *			list is empty. This is initialized to zero
 *			(first value) when the list is created.
 * DBR.currentAttrValue
 *			A copy of the current attribute value. The
 *			Data Display Manager uses the handle to
 *			determine what data to display.
 */
#include "DTSCatalogPeek.h"
#pragma segment MainCode

#define VALUE_LIST		(DBR.attributeValueList)
#define VALUE_BOUNDS	((**VALUE_LIST).dataBounds)
static const AttributeHeaderInfo 	gNullAttribute;		/* Null datum	*/

/*
 * Local function prototypes
 */
Boolean								GetNthAttributeValue(
		register CatalogBrowserPtr		dbp,
		short							cellIndex,
		Boolean							setSelect,
		AttributeHeaderInfoPtr			result
	);

/*
 * ClearAttributeValueList
 * This is called when a new directory record or attribute type
 * is selected to clear the attribute value list. Note that it
 * must read all cells to delete the data handle. The attribute
 * data display will also be cleared.
 */
void
ClearAttributeValueList(
		register CatalogBrowserPtr		dbp
	)
{
		AttributeHeaderInfo					info;
		short								i;
		
		FailNIL(VALUE_LIST);
		/*
		 * Delete the data handles that are stored in the
		 * attribute value list item, then delete the list
		 * contents.
		 */
		for (i = 0; i < VALUE_BOUNDS.bottom; i++) {
			GetNthAttributeValue(dbp, i, FALSE, &info);
			ForgetHandle(info.dataHandle);
		}
		LDelRow(0, 0, VALUE_LIST);
		DBR.currentAttributeValueIndex = -1;
		DBR.currentAttrValue = gNullAttribute;
		ClearAttributeDataDisplay(dbp);
}

/*
 * DoClickInAttributeValueList
 * The user clicked in the attribute value list.
 * Get the new selectionm, ignoring double-clicks.
 * If the selection changed, clear the attribute data
 * display and display the new data.
 */
void
DoClickInAttributeValueList(
		register CatalogBrowserPtr		dbp
	)
{
		Cell				theCell;
		
		LClick(DBR.penPt, EVENT.modifiers, VALUE_LIST);
		theCell = LLastClick(VALUE_LIST);
		if (PtInRect(theCell, &VALUE_BOUNDS)
		 && theCell.v != DBR.currentAttributeValueIndex) {
			ClearAttributeDataDisplay(dbp);
			SelectNthAttributeValue(dbp, theCell.v);
		}
}

/*
 * SelectNthAttributeValue
 * Extract the current attribute value and display its contents.
 */
void
SelectNthAttributeValue(
		register CatalogBrowserPtr		dbp,
		short							index
	)
{
		if (GetNthAttributeValue(dbp, index, TRUE, &DBR.currentAttrValue)) {
			DBR.currentAttributeValueIndex = index;
			DisplayCurrentAttributeData(dbp);
		}
}

/*
 * AddAttributeValue
 * Store the new attribute at the end of the list. A "real"
 * program would sort it or something. Each list element contains
 * the fixed length attribute value information plus a handle
 * to the variable-length data.
 */
void
AddAttributeValue(
		register CatalogBrowserPtr		dbp,
		const Attribute					*attributePtr
	)
{
		Cell						theCell;
		AttributeHeaderInfo			info;
#define ATTR	(*attributePtr)

		/*
		 * An attribute consists of a fixed-length
		 * header containing some tags, a longword
		 * data length, and a pointer to some bytes.
		 * The list cell holds the fixed-length
		 * data, while the variable-length data is
		 * stored in a Handle.
		 */
		info.attributeType = ATTR.attributeType;
		info.cid = ATTR.cid;
		info.tag = ATTR.value.tag;
		FailOSErr(
			PtrToHand(
				ATTR.value.bytes,
				&info.dataHandle,
				ATTR.value.dataLength
			)
		);
		theCell.v = LAddRow(1, VALUE_BOUNDS.bottom,VALUE_LIST);
		FailMemError();
		theCell.h = 0;
		LSetCell((Ptr) &info, sizeof info, theCell, VALUE_LIST);
		FailMemError();
}

/*
 * GetNthAttributeValue
 * Select a specific attribute value and display its contents.
 */
Boolean
GetNthAttributeValue(
		register CatalogBrowserPtr		dbp,
		short							cellIndex,
		Boolean							setSelect,
		AttributeHeaderInfoPtr			result
	)
{
		short						dataLen;
		Cell						theCell;
		Boolean						gotOne;

		SetPt(&theCell, 0, cellIndex);
		if (PtInRect(theCell, &VALUE_BOUNDS) == FALSE) {
		 	*result = gNullAttribute;
		 	gotOne = FALSE;
		}
		else {
			dataLen = sizeof (AttributeHeaderInfo);
			LGetCell(result, &dataLen, theCell, VALUE_LIST);
			if (setSelect)
				LSetSelect(TRUE, theCell, VALUE_LIST);
			gotOne = TRUE;
		}
		return (gotOne);
}
		
