/*									BalloonHelp.c							*/
/*
 * DTS Catalog Peek
 * BalloonHelp.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * BalloonHelp draws an appropriate string in the window
 * rectangles. We don't care about disabled items (yet),
 * however it would be reasonable to add a segment that
 * understood whether items had content, and whether radio
 * buttons were selected. We don't add a specific help menu item.
 *
 * Snippit alert: all you need to use this in another application
 * is a list of balloon-able window rectangles and, preferably,
 * an understanding of the rectangle state. While this function
 * uses a static rectangle definition and does not recognize
 * state, it is quite easy to extend it to more dynamic window
 * definitions.
 *
 * Globals:
 *	gSystemHasBalloonHelp
 *			Gestalt sets this TRUE when the application starts
 *			if balloon help is present on this machine.
 *	gLastBalloonItem
 *			A static variable that remembers which item the
 *			mouse is currently selecting. Note that the
 *			first window item must be non-zero.
 *	gLastBalloonBrowser
 *			A static variable that remembers which window the
 *			mouse is currently selecting. NULL if none is
 *			selected.
 */
#include "DTSCatalogPeek.h"
#include <Balloons.h>
#pragma segment SupportCode

/*
 * These "globals" are needed to determine when the
 * mouse moves from one balloon item (or window) to
 * a new item or window.
 */
static short					gLastBalloonItem;
static CatalogBrowserPtr		gLastBalloonBrowser;

/*
 * InitializeBalloonHelp is called when the application starts to add a
 * Help menu item to the balloon menu. Note that we never reference
 * gHelpMenu and gHelpMenuItem - they are needed only if we change to
 * a non-modal help menu: we then check the help menu item when the
 * help window is foremost.
 */
void
InitializeBalloonHelp(void)
{
		OSErr				status;
		StringHandle		menuName;
		
		if (gSystemHasBalloonHelp) {
			gLastBalloonItem = 0;
			gLastBalloonBrowser = NULL;
			menuName = GetString(HELP_Browser);
			status = HMGetHelpMenuHandle(&gHelpMenu);
			if (menuName != NULL && status == noErr && gHelpMenu != NULL) {
				HLock((Handle) menuName);
			 	gHelpMenuItem = CountMItems(gHelpMenu) + 1;
				AppendMenu(gHelpMenu, *menuName);
				HUnlock((Handle) menuName);
				ReleaseResource((Handle) menuName);
			}
		}
}

/*
 * BalloonHelp
 * Manage balloon help.
 */
void
BalloonHelp(
		register CatalogBrowserPtr		dbp
	)
{
		Point				mousePt;
		short				i;
		Rect				hotRect;
		HMMessageRecord		helpMsg;
		OSErr				status;
		short				helpItemIndex;
		Boolean				inItemRect;

		if (gSystemHasBalloonHelp && HMGetBalloons()) {
			/*
			 * Help is present and enabled.
			 * Our window is active and is the active port.
			 */
			GetMouse(&mousePt);
			inItemRect = FALSE;
			if (PtInRect(mousePt, &WINDOW->portRect)) {
				for (i = kFirstItemRect; i <= kLastItemRect; i++) {
					if (PtInRect(mousePt, &RECT(i))) {
						/*
						 * The mouse is in one of our items.
						 */
						if (i != gLastBalloonItem
						 || dbp != gLastBalloonBrowser) {
						 	/*
						 	 * The mouse has moved to new balloon.
						 	 */
						 	hotRect = RECT(i);
						 	LocalToGlobal((Point *) &topLeft(hotRect));
						 	LocalToGlobal((Point *) &botRight(hotRect));
						 	SetPt(
						 		&mousePt,
						 		hotRect.left + width(hotRect) / 2,
						 		hotRect.top + height(hotRect) / 2
						 	);
						 	/*
						 	 * At this point in the program, we must to examine
						 	 * what the mouse is pointing to (is it enabled or
						 	 * selected, for example), so we can be more selective
						 	 * in our choice of a message.
						 	 */
						 	helpItemIndex = GetHelpItemIndex(dbp, i);
						 	status = HMExtractHelpMsg(
						 				kHMDialogResType,
						 				HELP_Browser,
						 				i,					/* Item id			*/
						 				helpItemIndex,		/* Item state		*/
						 				&helpMsg
						 			);
							if (status != noErr)
								break;
							status = HMShowBalloon(
										&helpMsg,
										mousePt,
										&hotRect,
										NULL,				/* No tip Proc		*/
										0, 0,				/* Default balloon	*/
										kHMRegularWindow	/* Normal update	*/
									);
							if (status != noErr)
								break;
						 	gLastBalloonItem = i;
							gLastBalloonBrowser = dbp;
						}				/* Mouse is in a new balloon			*/
						inItemRect = TRUE;		/* The mouse is in a balloon	*/
						break;					/* Exit for loop				*/
					}							/* Mouse is in a window item	*/
				}								/* For each window item...		*/
			}									/* Mouse is in our window		*/
			if (inItemRect == FALSE)
				gLastBalloonBrowser = NULL;		/* No longer in a balloon		*/
		}										/* Help is present and enabled	*/
}
