/*							CatalogDisplayFormatter.c							*/
/*
 * DTS Catalog Peek
 * CatalogDisplayFormatter.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions manage the display list that shows the current directory
 * expanded as far as reasonably possible, and then some.
 *
 * The logic of this module follows the logic of UnpackSDPPanelSelection.c -- it
 * completely unpacks the current selection using, for convenienct, the same DBR
 * variables as UnpackSDPPanelSelection. The unpacked selection is organized as a
 * linked list with more detailed information "deeper" into the list. I.e., the
 * top level has the directory type, record id top-level, extension top-level, and
 * raw directory content top-level. The user can click on a twist-down button to
 * reveal the record id content, etc.
 *
 * All functions that return TwistDownHdl's return the starting element of the
 * the sublist that they create. This will be stored into the ELEM.subList
 * variable in the list's parent node. 
 *
 * Globals
 *	(All of the directory-content globals are also accessed).
 */
#define EXPAND_SUBLISTS	1
#include "DTSCatalogPeek.h"
#pragma segment MainCode

#ifndef EXPAND_SUBLISTS
#define EXPAND_SUBLISTS	0
#endif

TwistDownHdl						DisplayExtensionValue(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayPackedRLIPtr(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayLocalRecordID(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayPackedRLIPtr(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayUnpackedRLI(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayPathNames(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);
TwistDownHdl						DisplayLocalRecordID(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	);

#define ELEM		(**thisElement)
#if EXPAND_SUBLISTS
#undef EXPAND_SUBLISTS
#define EXPAND_SUBLISTS	(ELEM.flag.showSublist = TRUE)
#else
#define EXPAND_SUBLISTS	/* Nothing */
#endif
#define HAS_TWISTDOWN	(ELEM.flag.hasTwistDown = TRUE);

/*
 * DisplayDSSpecType
 * Format the directory spec. type. This is the first one to be called,
 * so we actually can ignore previousIndex.
 *
 * DisplayDSSpecType is a leaf node, called from MakeCatalogDisplayList.
 */
TwistDownHdl
DisplayDSSpecType(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		Str255						work;
		
		work[0] = 0;
		DBR.specType = OCEGetDSSpecInfo(&DBR.theDSSpec);
		AppendOSType(work, DBR.specType);
		AppendPascalString(work, "\p: ");
		switch (DBR.specType) {
		case kOCEInvalidDSSpec:
			AppendPascalString(work, "\pInvalid DSSpec type");
			break;
		case kOCEDirsRootDSSpec:
			AppendPascalString(work, "\pRoot of all directories");
			break;
		case kOCEDirectoryDSSpec:
			AppendPascalString(work, "\pCatalog");
			break;
		case kOCEDNodeDSSpec:
			AppendPascalString(work, "\pCatalog node");
			break;
		case kOCERecordDSSpec:
			AppendPascalString(work, "\pRecord");
			break;
		case kOCEentnDSSpec:
			AppendPascalString(work, "\pExtension type is 'entn'");
			break;
		case kOCENOTentnDSSpec:
			AppendPascalString(work, "\pExtension type is not 'entn'");
			break;
		default:
			AppendPascalString(work, "\pUnknown Catalog Spec. type");
			break;
		}
		return (StoreElementString(previousElement, indentLevel, work));
}

/*
 * DisplayRawDSSpecPtr
 * Display the unformatted DSSpec.
 * DisplayDSSpecPtr is called from MakeCatalogDisplayList. It calls
 * DisplayDataBlock to format the data.
 */
TwistDownHdl
DisplayRawDSSpecPtr(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		Str255						work;
		
		if (DBR.selectedDSSpec == NULL) {
			thisElement = StoreElementMessage(
						previousElement,
						indentLevel,
						"\pNULL packedDSSpec pointer"
					);
		}
		else {
			work[0] = 0;
			AppendPascalString(work, "\pPacked DSSpec (");
			AppendUnsigned(work, DBR.selectedDSSpecSize);
			AppendPascalString(work, "\p bytes)");
			thisElement = StoreElementString(
						previousElement,
						indentLevel,
						work
					);
			HAS_TWISTDOWN;
			EXPAND_SUBLISTS;
			ELEM.subElement = DisplayDataBlock(
								dbp,
								NULL,
								indentLevel + 1,
								(Ptr) DBR.selectedDSSpec,
								DBR.selectedDSSpecSize
							);
		}
		return (thisElement);
}

/*
 * DisplayRecordID
 * Display the unpacked directory specifier.
 *
 * DisplayRecordID is called from MakeCatalogDisplayList. If there is a RecordID
 * structure, it is unpacked as a sub-list, otherwise, this is a leaf node.
 */
TwistDownHdl
DisplayRecordID(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		Str255						work;
		
		work[0] = 0;
		if (DBR.theDSSpec.entitySpecifier == NULL) {
			thisElement = StoreElementMessage(
						previousElement,
						indentLevel,
						"\pNull Record ID"
					);
		}
		else {
			AppendPascalString(work, "\pRecord ID.");
			thisElement = StoreElementString(
						previousElement,
						indentLevel,
						"\pRecord ID"
					);
			ELEM.subElement = DisplayPackedRLIPtr(dbp, NULL, indentLevel + 1);
			HAS_TWISTDOWN;
			EXPAND_SUBLISTS;
			(void) DisplayLocalRecordID(dbp, ELEM.subElement, indentLevel + 1);
		}
		return (thisElement);	
}

/*
 * DisplayExtension
 * Format the extension type, size, and value from the current DSSpec. The
 * extension value, if present, is expanded as a sub-list.
 *
 * DisplayExtension is called from MakeCatalogDisplayList.
 */
TwistDownHdl
DisplayExtension(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		Str255						work;
		OSType						extensionType;
		unsigned short				extensionSize;
		Ptr							extensionValue;
				
		work[0] = 0;
		extensionType = DBR.theDSSpec.extensionType;
		extensionSize = DBR.theDSSpec.extensionSize;
		extensionValue = DBR.theDSSpec.extensionValue;
		AppendPascalString(work, "\pExtension ");
		AppendOSType(work, extensionType);
		AppendPascalString(work, "\p type, ");
		AppendUnsigned(work, extensionSize);
		AppendPascalString(work, "\p size, ");
		if (extensionValue == NULL)
			AppendPascalString(work, "\p NULL value");
		else {
			AppendHexLeadingZeros(
				work,
				(unsigned long) extensionValue,
				sizeof (Ptr) * 2
			);
			AppendPascalString(work, "\p value.");
		}
		thisElement = StoreElementString(previousElement, indentLevel, work);
		if (extensionType == kOCEentnDSSpec		/* 'entn'	*/
		 && extensionValue != NULL
		 && extensionSize > 0) {
			ELEM.subElement = DisplayExtensionValue(dbp, NULL, indentLevel + 1);
			HAS_TWISTDOWN;
			EXPAND_SUBLISTS;
		}
		return (thisElement);
}

/*
 * DisplayExtensionValue
 * Unpack the extension value (the caller checked that there really is one).
 * DisplayExtensionValue is called from DisplayExtension. It is a leaf node.
 */
TwistDownHdl
DisplayExtensionValue(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		return (
			DisplayDataBlock(
				dbp,
				previousElement,
				indentLevel,
				DBR.theDSSpec.extensionValue,
				DBR.theDSSpec.extensionSize
		));
}

/*
 * DisplayPackedRLIPtr
 * Unpack and display the RLI (record location identifier) stored in
 * SDP_INFO_RecordID.rli.
 *
 * DisplayPackedRLIPtr is called from DisplayRecordID. It unpacks the RLI,
 * displays the contents, and calls DisplayPackedPathName to display the RLI path.
 */ 
TwistDownHdl
DisplayPackedRLIPtr(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		Str255						work;
				
		work[0] = 0;
		if (OCEValidPackedRLI(SDP_INFO_RecordID.rli) == FALSE) {
			AppendPascalString(work, "\pInvalid PackedRLI");
			thisElement = StoreElementString(previousElement, indentLevel, work);
		}
		else {
			OCEUnpackRLI(SDP_INFO_RecordID.rli, &DBR.theDSSpecRLI);
			AppendPascalString(work, "\pRLI, dNodeNumber = ");
			AppendUnsigned(work, DBR.theDSSpecRLI.dNodeNumber);
			thisElement = StoreElementString(previousElement, indentLevel, work);
			ELEM.subElement = DisplayUnpackedRLI(dbp, NULL, indentLevel + 1);
			HAS_TWISTDOWN;
			EXPAND_SUBLISTS;
		}
		return (thisElement);
}

/*
 * DisplayUnpackedRLI
 * This displays the fixed part of the unpacked RLI. If there is a path, it calls
 * DisplayPathNames to format the path name strings.
 */
TwistDownHdl
DisplayUnpackedRLI(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		TwistDownHdl				firstElement;
		Str255						work;
		unsigned short				nodeNameCount;
#define DIRNAME	(DBR.theDSSpecRLI.directoryName)
#define DISCRIM	(DBR.theDSSpecRLI.discriminator)
#define PATH	(DBR.theDSSpecRLI.path)

		if (DIRNAME == NULL) {
			thisElement = StoreElementMessage(
						previousElement,
						indentLevel,
						"\pNull Catalog Name"
					);
		}
		else {
			thisElement = StoreElementRString(
						previousElement,
						indentLevel,
						(RStringPtr) DIRNAME
					);
		}
		firstElement = thisElement;
		work[0] = 0;
		AppendPascalString(work, "\pDirDiscriminator: ");
		AppendOSType(work, DISCRIM.signature);
		AppendPascalString(work, "\p discriminator, ");
		AppendHexLeadingZeros(work, DISCRIM.misc, sizeof DISCRIM.misc);
		AppendPascalString(work, "\p misc");
		thisElement = StoreElementString(thisElement, indentLevel, work);
		if (OCEValidPackedPathName(PATH) == FALSE) {
			thisElement = StoreElementMessage(
						thisElement,
						indentLevel,
						"\pInvalid RLI.directoryName.path"
					);
		}
		else {
			work[0] = 0;
			AppendPascalString(work, "\pRLI.directoryName.path, ");
			nodeNameCount = OCEDNodeNameCount(PATH);
			AppendUnsigned(work, (unsigned long) nodeNameCount);
			AppendPascalString(work, "\p directory node names");
			thisElement = StoreElementString(thisElement, indentLevel, work);
			if (nodeNameCount != 0) {
				ELEM.subElement = DisplayPathNames(dbp, NULL, indentLevel + 1);
				HAS_TWISTDOWN;
				EXPAND_SUBLISTS;
			}
		}
		return (firstElement);
#undef DIRNAME
#undef DISCRIM
#undef PATH
}

/*
 * DisplayPathNames
 * Expand the directory path names and build a list of
 * path names. This is a leaf node.
 */
TwistDownHdl
DisplayPathNames(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		TwistDownHdl				firstElement;
		Str255						work;
		unsigned short				nodeNameCount;
		RStringPtr					*partsPtr;
		OSErr						nParts;
		short						i;
#define DIRNAME	(DBR.theDSSpecRLI.directoryName)
#define DISCRIM	(DBR.theDSSpecRLI.discriminator)
#define PATH	(DBR.theDSSpecRLI.path)

		nodeNameCount = OCEDNodeNameCount(PATH);
		/*
		 * This test catches a bug in early alpha versions of AOCE.
		 */
		if (nodeNameCount > 1000) {
			work[0] = 0;
			AppendPascalString(work, "\pBogus node name count: ");
			AppendUnsigned(work, nodeNameCount);
			firstElement = StoreElementMessage(
						previousElement,
						indentLevel,
						work
					);
		}
		else {
			firstElement = NULL;
			partsPtr = NULL;
			TRY {
				partsPtr = (RStringPtr *) NewPtrClear(
							sizeof (RStringPtr) * nodeNameCount
						);
				FailNIL(partsPtr);
				nParts = OCEUnpackPathName(PATH, partsPtr, nodeNameCount);
				if (nParts != nodeNameCount) {
					work[0] = 0;
					AppendPascalString(work, "\pExpected ");
					AppendUnsigned(work, nodeNameCount);
					AppendPascalString(work, "\p parts, but found ");
					AppendUnsigned(work, nParts);
					thisElement = StoreElementString(
								previousElement,
								indentLevel,
								work
							);
					previousElement = thisElement;
					firstElement = thisElement;
				}
				for (i = 0; i < nParts; i++) {
					thisElement = StoreElementRString(
								previousElement,
								indentLevel,
								partsPtr[i]
							);
					if (firstElement == NULL)
						firstElement = thisElement;
					previousElement = thisElement;
				}
			}
			CATCH {
				NO_PROPAGATE;
			}
			ENDTRY;
			ForgetPtr(partsPtr);
		}
		return (firstElement);
#undef DIRNAME
#undef DISCRIM
#undef PATH
}

/*
 * DisplayLocalRecordID
 * Unpack and display the local record identifier stored in
 * SDP_INFO_RecordID.local.
 *
 * DisplayLocalRecordID is called from DisplayRecordID. It displays
 * the creation id and creates a sub-list to display the recordName and
 * recordType RStrings.
 */
TwistDownHdl
DisplayLocalRecordID(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					previousElement,
		short							indentLevel
	)
{
		TwistDownHdl				thisElement;
		TwistDownHdl				firstElement;
		Str255						work;
#define LOCAL	(SDP_INFO_RecordID.local)
#define CID		(LOCAL.cid)

		work[0] = 0;
		AppendPascalString(work, "\pLocalRecordID. Creation id = ");
		AppendHexLeadingZeros(
			work,
			(unsigned long) CID.source,
			sizeof (unsigned long) * 2
		);
		AppendChar(work, ' ');
		AppendHexLeadingZeros(
			work,
			(unsigned long) CID.seq,
			sizeof (unsigned long) * 2
		);
		thisElement = StoreElementString(previousElement, indentLevel, work);
		firstElement = thisElement;
		HAS_TWISTDOWN;
		EXPAND_SUBLISTS;
		if (LOCAL.recordName == NULL) {
			ELEM.subElement = StoreElementMessage(
						NULL,
						indentLevel + 1,
						"\pNULL Local Record Name"
					);
		}
		else {
			ELEM.subElement = StoreElementRString(
						NULL,
						indentLevel + 1,
						LOCAL.recordName
					);
		}
		if (LOCAL.recordType == NULL) {
			(void) StoreElementMessage(
						ELEM.subElement,
						indentLevel + 1,
						"\pNULL Local Record Type"
					);
		}
		else {
			(void) StoreElementRString(
						ELEM.subElement,
						indentLevel + 1,
						LOCAL.recordType
					);
		}
		return (firstElement);
#undef LOCAL
#undef CID
}

