/*							CatalogDisplayManager.c								*/
/*
 * DTS Catalog Peek
 * CatalogDisplayManager.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions manage the display list that shows the current directory
 * expanded as far as reasonably possible, and then some. Functions in this file
 * manage the list itself. The formatters are elsewhere.
 *
 * The list consists of Lisp-style linked lists that are managed by functions
 * in the TwistDownList package. In particular, each list element contains
 * two pointers:
 *		nextElement corresponds to a CDR pointer. It locates the next element
 *			on this level.
 *		subElement corresponds to a CAR pointer. It locates the first element
 *			in a child list.
 * MakeTwistDownElement corresponds to the CONS function. It creates a new list
 * element that is pointed to by the previousElement parameter. I.e., it creates a
 * new list element and sets CAR(previousElement) to point to this element
 * (assuming that index is not the magic "at list head" NULL value). It then
 * returns the TwistDownHandle that contains the new element.
 *
 * Note also that the showSubList flag is set TRUE if the sublist is to be shown on
 * the display.
 *
 * Ok, stop whining about CAR and CDR: I programmed an IBM 7090, so I actually know
 * what CAR and CDR mean. And, I still have the Lisp 1.5 manuals, so there!
 */
#include "DTSCatalogPeek.h"
#pragma segment MainCode

/*
 * If the user selects a directory and chooses to display the directory list
 * (rather than the record data content), build the complete list here. The process
 * is slightly messy: each "chunk" of data consists of a list of items. Some
 * elements have sub-lists.
 */
void
MakeCatalogDisplayList(
		register CatalogBrowserPtr		dbp
	)
{
		TwistDownHdl					thisElement;
#define SET_FIRST	do {								\
			if (DBR.directoryListHead == NULL)			\
				DBR.directoryListHead = thisElement;	\
		} while (0)

		if (DBR.showCatalog) {
			ClearCatalogDataDisplay(dbp);
			if (DBR.selectedDSSpec != NULL) {
				/*
				 * Display the directory.
				 */
				thisElement = DisplayDSSpecType(dbp, NULL, 0);
				SET_FIRST;
				thisElement = DisplayRecordID(dbp, thisElement, 0);
				SET_FIRST;
				thisElement = DisplayExtension(dbp, thisElement, 0);
				SET_FIRST;
				thisElement = DisplayRawDSSpecPtr(dbp, thisElement, 0);
				SET_FIRST;
			}
			CreateVisibleList(DBR.directoryDisplayList, DBR.directoryListHead);
		}
#undef SET_FIRST
}

void
ClearCatalogDataDisplay(
		register CatalogBrowserPtr		dbp
	)
{
		if (DBR.directoryListHead != NULL) {
			LDelRow(0, 0, DBR.directoryDisplayList);
			DisposeTwistDownHdl(DBR.directoryListHead);
			DBR.directoryListHead = NULL;
		}
}

/*
 * Display an unformatted block of text. This is a leaf node.
 * Note: if the data is derived from a handle, the handle
 * should be locked.
 */
TwistDownHdl
DisplayDataBlock(
		register CatalogBrowserPtr		dbp,
		TwistDownHdl					thisElement,	/* Linkage				*/
		short							indentLevel,
		Ptr								dataPtr,
		unsigned short					dataLength
	)
{
#pragma unused (dbp)
		TwistDownHdl				firstElement;
		Str255						work;
		unsigned short				i;
		short						nBytesToDraw;
		short						offset;
		unsigned short				c;
		
		firstElement = NULL;
		for (offset = 0; offset < dataLength; offset += kBytesPerLine) {
			if ((offset + kBytesPerLine) <= dataLength)
				nBytesToDraw = kBytesPerLine;
			else {
				nBytesToDraw = dataLength - offset;
			}
			work[0] = 0;
			AppendHexLeadingZeros(
				work,
				offset,
				sizeof (unsigned short) * 2
			);
			AppendChar(work, ':');
			for (i = 0; i < nBytesToDraw; i++) {
				AppendChar(work, ' ');
				AppendHexLeadingZeros(work, dataPtr[offset + i], 2);
			}
			for (; i < kBytesPerLine; i++)
				pstrcat(work, "\p   ");
			AppendChar(work, ' ');
			for (i = 0; i < nBytesToDraw; i++) {
				c = dataPtr[offset + i];
				if (c < ' ')
					c = '.';
				AppendChar(work, c);
			}
			thisElement = StoreElementFormattedString(
							thisElement,
							indentLevel,
							kElementCode,
							work
						);
			if (firstElement == NULL)
				firstElement = thisElement;
		}
		return (firstElement);
}


/*
 * Store an RString in the indicated node.
 */
TwistDownHdl
StoreElementRString(
		TwistDownHdl					previousElement,
		short							indentLevel,
		const RStringPtr				srcRStringPtr
	)
{
		TwistDownHdl				thisElement;
		
		if (srcRStringPtr == NULL) {
			thisElement = StoreElementMessage(
						previousElement,
						indentLevel,
						"\p? Null RString ?"
					);
		}
		else if (OCEValidRString(srcRStringPtr, kOCEGenericSensitive) == FALSE) {
			thisElement = StoreElementMessage(
						previousElement,
						indentLevel,
						"\p? Invalid RString ?"
					);
		}
		else {
			thisElement = StoreElementDatum(
						previousElement,
						indentLevel,
						kElementRString,
						RStringLength(srcRStringPtr),
						(const Ptr) srcRStringPtr
					);
		}
		return (thisElement);
}

/*
 * Store a string in the indicated node.
 */
TwistDownHdl
StoreElementMessage(
		TwistDownHdl					previousElement,
		short							indentLevel,
		ConstStr255Param				work
	)
{
		TwistDownHdl					thisElement;
		
		thisElement = StoreElementFormattedString(
					previousElement,
					indentLevel,
					kElementMessage,
					work
				);
		return (thisElement);
}


/*
 * Store a normal string
 */
TwistDownHdl
StoreElementString(
		TwistDownHdl					previousElement,
		short							indentLevel,
		ConstStr255Param				work
	)
{
		TwistDownHdl					thisElement;
		
		thisElement = StoreElementFormattedString(
					previousElement,
					indentLevel,
					kElementString,
					work
				);
		return (thisElement);
}

/*
 * Store a Str255 string in some format.
 */
TwistDownHdl
StoreElementFormattedString(
		TwistDownHdl					previousElement,
		short							indentLevel,
		ElementFormat					elementFormat,
		ConstStr255Param				work
	)
{
		TwistDownHdl					thisElement;

		thisElement = StoreElementDatum(
					previousElement,
					indentLevel,
					elementFormat,
					work[0],
					(const Ptr) &work[1]
				);
		return (thisElement);
}

/*
 * Store one of our elements.
 */
TwistDownHdl
StoreElementDatum(
		TwistDownHdl					previousElement,
		short							indentLevel,
		ElementFormat					elementFormat,
		unsigned short					dataLength,
		const Ptr						dataPtr
	)
{
		OSErr							status;
		unsigned short					elementLength;
		TwistDownHdl					thisElement;
		ElementPtr						elementPtr;
		

		elementLength = ELEM_HEADER_SIZE + dataLength;
		status = MakeTwistDownElement(
					previousElement,
					indentLevel,
					elementLength,
					NULL,				/* We fill it in		*/
					&thisElement
				);
		FailOSErr(status);
		/* HLock(thisElement);			-- not needed			*/
		elementPtr = (ElementPtr) (**thisElement).data;
		(*elementPtr).header.format = elementFormat;
		BlockMove(dataPtr, (*elementPtr).data, dataLength);
		/* HUnlock(thisElement);		-- not needed			*/
		return (thisElement);
}


