/*								DataDisplayManager.c							*/
/*
 * DTS Catalog Peek
 * DataDisplayManager.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions manage the display list that shows the actual content of the
 * current attribute value.
 *
 * Global values
 *	DBR.showCatalog
 *			Boolean: if TRUE, the user wants to display the expanded directory,
 *			if FALSE, the user is displaying the record contents.
 *	DBR.dataDisplayList
 *			This list contains the formatted record text.
 *	DBR.directoryDisplayList
 *			This list contains the expanded directory. Either DBR.dataDisplayList
 *			of DBR.directoryDisplayList is visible, not both.
 *	DBR.currentAttrValue.dataHandle
 *			This handle contains the record data that is to be displayed.
 */
#include "DTSCatalogPeek.h"
#pragma segment MainCode

/*
 * Local function prototypes.
 */
static void							DrawOneDataDisplayLine(
		register CatalogBrowserPtr		dbp,
		Ptr								dataPtr,
		Size							offset,
		short							nBytesToDraw
	);
static void
AddDisplayAttributeDataRow(
		register CatalogBrowserPtr		dbp,
		short							rowNumber
	);

/*
 * ClearAttributeDataDisplay
 * Do nothing if the user is displaying the formatted directory. Else, empty the
 * content.
 */
void
ClearAttributeDataDisplay(
		register CatalogBrowserPtr		dbp
	)
{
		if (DBR.showCatalog == FALSE && DBR.dataDisplayList != NULL) {
		 	EraseRect(&RECT(kInRecordLengthText));
		 	ValidRect(&RECT(kInRecordLengthText));
			LDelRow(0, 0, DBR.dataDisplayList);
		}
}

/*
 * DoClickInDataDisplayList
 * If the user is displaying the formatted directory, call DoTwistDownClick and
 * ExpandOrCollapseTwistDownList to manage the display. Otherwise, just call
 * LClick to scroll the list.
 */
void
DoClickInDataDisplayList(
		register CatalogBrowserPtr		dbp
	)
{
		Cell						theCell;
		TwistDownClickState			clickState;
		
		if (DBR.showCatalog) {
			clickState = DoTwistDownClick(
						DBR.directoryDisplayList,
						&EVENT,
						&theCell
					);
		}
		else {
			/*
			 * Call LClick to do scrolling. However, the display list doesn't have
			 * anything selectable.
			 */
			LClick(DBR.penPt, EVENT.modifiers, DBR.dataDisplayList);
		}
}

/*
 * DisplayCurrentAttributeData
 * The current data has changed. Format it and store the
 * formatted text into the list.
 */
void
DisplayCurrentAttributeData(
		register CatalogBrowserPtr		dbp
	)
{
		Size						dataLength;
		short						currentLines;
		short						i;
		
		if (DBR.showCatalog == FALSE) {
			if (DBR.currentAttrValue.dataHandle == NULL)
				dataLength = 0;
			else {
				dataLength = GetHandleSize(
								DBR.currentAttrValue.dataHandle
							);
			}
			currentLines =
					(dataLength + kBytesPerLine - 1) / kBytesPerLine;
			LDoDraw(FALSE, DBR.dataDisplayList);
			LDelRow(0, 0, DBR.dataDisplayList);		/* Delete current contents	*/
			LAddRow(currentLines, 0, DBR.dataDisplayList);
			for (i = 0; i < currentLines; i++)
				AddDisplayAttributeDataRow(dbp, i);
			LDoDraw(TRUE, DBR.dataDisplayList);
			LUpdate((**DBR.dataDisplayList).port->visRgn, DBR.dataDisplayList);
			DrawDataRecordLength(dbp);
		}
}

/*
 * DrawDataRecordLength
 * Draw the length of the attribute data handle.
 */
void
DrawDataRecordLength(
		register CatalogBrowserPtr		dbp
	)
{
		Str255						work;
		Size						dataLength;
		
		EraseRect(&RECT(kInRecordLengthText));
		if (DBR.currentAttrValue.dataHandle != NULL
		 && DBR.showCatalog == FALSE) {
			dataLength = GetHandleSize(
							DBR.currentAttrValue.dataHandle
						);
			work[0] = 0;
			AppendUnsigned(work, dataLength);
			AppendPascalString(work,
				(dataLength == 1)
					? "\p byte record length"
					: "\p bytes record length"
			);
			TextFont(DBR.labelFont.fontNumber);
			TextSize(DBR.labelFont.fontSize);
			DrawWideString(work, &RECT(kInRecordLengthText), normal);
		}
}

/*
 * AddDisplayAttributeDataRow
 * Add one line of formatted data to the display list.
 */
static void
AddDisplayAttributeDataRow(
		register CatalogBrowserPtr		dbp,
		short							rowNumber
	)
{
#pragma unused (dbp)
		Str255					work;
		short					i;
		unsigned short			c;
		Cell					theCell;
		Ptr						dataPtr;
		short					nBytesToDraw;
		short					offset;
		Handle					theHandle;
		
		/*
		 * Note that we don't need to lock the data handle;
		 * but we do anyway because of rampant paranoia.
		 */
		theHandle = DBR.currentAttrValue.dataHandle;
		HLock(theHandle);
		dataPtr = *theHandle;
		offset = rowNumber * kBytesPerLine;
		if ((offset + kBytesPerLine) <= GetHandleSize(theHandle))
			nBytesToDraw = kBytesPerLine;
		else {
			nBytesToDraw = GetHandleSize(theHandle) - offset;
		}
		work[0] = 0;
		AppendHexLeadingZeros(work, offset, sizeof (Size) * 2);
		AppendChar(work, ':');
		for (i = 0; i < nBytesToDraw; i++) {
			AppendChar(work, ' ');
			AppendHexLeadingZeros(work, dataPtr[offset + i], 2);
		}
		for (; i < kBytesPerLine; i++)
			pstrcat(work, "\p   ");
		AppendChar(work, ' ');
		for (i = 0; i < nBytesToDraw; i++) {
			c = dataPtr[offset + i];
			if (c < ' ')
				c = '.';
			AppendChar(work, c);
		}
		HUnlock(theHandle);
		SetPt(&theCell, 0, rowNumber);
		LSetCell(&work[1], work[0], theCell, DBR.dataDisplayList);
		FailOSErr(MemError());
}

