/*								DecorateDisplay.c								*/
/*
 * DTS Catalog Peek
 * DecorateDisplay.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 * DecorateDisplay organizes the visual appearance of the display. It is called
 * when the display is created and resized.
 */
#include "DTSCatalogPeek.h"
#include <ToolUtils.h>
#include <FixMath.h>

#pragma segment MainCode

void								AdjustListBottom(
		register CatalogBrowserPtr		dbp,
		WindowFocus						index,
		short							cellHeight
	);
void								AdjustListTop(
		register CatalogBrowserPtr		dbp,
		WindowFocus						index,
		short							cellHeight
	);
void								RepositionPanel(
		register CatalogBrowserPtr		dbp,
		SDPPanelHandle					panelHandle
	);
void								RepositionList(
		ListHandle						theList,
		const Rect						*viewRect
	);
void								RepositionControl(
		register CatalogBrowserPtr		dbp,
		ControlHandle					theControl,
		WindowFocus						index
	);

#define kHSeparator				2			/* Generic item separator			*/
#define kVSeparator				2			/* Generic item separator			*/
#define kListVSeparator			(kHSeparator + 2) /* List bottom to list top	*/

/*
 * These values configure the display. The left edge of the record name is
 * repositioned as the window grows and shrinks. It's precise position is the
 * guidepost for all horizontal positioning. Note: all values are absolute:
 * the code body decides whether something moves to the left or up (negative)
 * or to the right and down (positive).
 */
#define kRecordNameLeft			234			/* Left edge of record name text	*/
#define kRecordNameTop			2			/* Top edge of record name and type	*/
#define kRecordNameHSeparator	2			/* Between record name and type		*/
#define kRecordNameHeight		18			/* Record name and type box height	*/
#define kRecordTypeRightIndent	4			/* Record type from right edge		*/
/*
 * These lists occupy the right-hand side of the display. Their heights are
 * adjusted as the window height changes. The vertical separation is fixed.
 * Note that the sum of the heights must fit in the window!
 */
#define kListRightIndent		(-1)		/* Right edge of lists hangs out	*/
#define kTypeListHeight			57			/* Height of Type list box			*/
#define kValueListHeight		57			/* Height of Value list box			*/
#define kErrorLogHeight			88			/* Height of Error list box			*/
#define kDataListHeight			88			/* Height of Data list box			*/
/*
 * Configure the spinner
 */
#define kSpinnerLeft			(kHSeparator) /* Spinner fixed left position	*/
#define	kSpinnerTop				(kVSeparator) /* Spinner fixed top position		*/
#define	kSpinnerSize			16			/* Spinner height and width			*/
/*
 * Position the Browser Popup - we should use some function of font height to
 * configure the browser. Also, we should look at the control size. The browser
 * popup is fixed at the top and centered over the SDP Panel. The left and right
 * edges of the browser popup are inset kBrowserPopupInset from the edges of
 * the SDP Panel. This inset must be big enough to clear the spinner.
 */
#define kBrowserTop				(kHSeparator) /* Browser fixed top position		*/
#define kBrowserHeight			16			/* Browser rectangle height			*/
#define kBrowserPopupInset		48			/* Browser popup horizontal inset	*/
/*
 * Position the SDP Panel - height and width are calculated.
 */
#define kSDPLeft				7			/* From the left edge				*/
#define kSDPRight				16			/* From the Record Name field		*/
/*
 * Configure the icon display and Icon choice popup control.
 */
#define kSDPIconWidth			54			/* SDP icons (big and small)		*/
#define kSDPIconHeight			32			/* SDP icon height (enough for big)	*/
#define kDirIconWidth			32			/* CSAM icon width					*/
#define kDirIconHeight			32			/* CSAM icon height					*/
#define kIconPopupWidth			52			/* Icon popup control width			*/
#define kIconPopupHeight		16			/* Icon popup control height		*/
#define kIconPopupHSeparator	4			/* Icon popup right to record name	*/
#define kSDPIconHOffset			44			/* Icon popup left to SDP icon left	*/
#define kSDPIconVOffset			20			/* Icon popup top to SDP icon top	*/
#define kDirIconHOffset			16			/* Icon popup left to Dir Icon left	*/
#define kDirIconVOffset			4			/* Icon popup bot to Dir Icon top	*/
/*
 * Configure the radio buttons and the record length text item: these all align
 * to the SDP Panel's left edge.
 */
#define kRadioButtonHeight		12			/* Button height					*/
#define kRadioButtonVSeparator	1			/* Between buttons					*/
#define kRecordLengthVSeparator	1			/* Cat. button bottom to text top	*/
#define kRecordLengthWidth		120			/* Record length text width			*/
#define kRecordLengthHeight		12			/* Record length text height		*/

/*
 * These macros compute a position or height as a fraction of the display size.
 * The values above are accurate for the minimum (Classic) display. As it grows
 * to fill larger displays, the contents grow so they occupy the same fraction
 * of the display.
 */
#define HorizontalRatio(v)	(FixRound(FixMul(Long2Fix(v), xRatio)))
#define VerticalRatio(v)	(FixRound(FixMul(Long2Fix(v), yRatio)))
#define SetHeight(i, v)		(RECT(i).bottom = RECT(i).top + VerticalRatio((v)))

/*
 * This repositions fixed-sized elements with respect to their top-left corner.
 * Positive offsets always move to the right and down.
 */
#define SetFixedShape(i, width, height) (			\
		RECT(i).right = RECT(i).left + width,		\
		RECT(i).bottom = RECT(i).top + height		\
	)
#define OffsetFromTop(offset, i)	(RECT(i).top + (offset))
#define OffsetFromLeft(offset, i)	(RECT(i).left + (offset))
#define OffsetFromRight(offset, i)	(RECT(i).right + (offset))
#define OffsetFromBottom(offset, i)	(RECT(i).bottom + (offset))

/*
 * This does all the work of repositioning the "normal" lists.
 */
#define SetListShape(thisList, aboveItem, thisHeight, cellHeight) do { \
		SetRect(												\
			&RECT(thisList),									\
			RECT(kInLocalRecordName).left,						\
			OffsetFromBottom(kListVSeparator, aboveItem),		\
			OffsetFromRight(-kListRightIndent, kWindowShape),	\
			0													\
		);														\
		SetHeight(thisList, thisHeight);						\
		AdjustListBottom(dbp, thisList, cellHeight);			\
	} while (0)


/*
 * DecorateDisplay is called when the window is created or when it changes size.
 * It computes all item rectangles. Note that DecorateDisplay may be called
 * before the display has been created. (In this case, it will not move any
 * controls, of course.)
 */ 
void
DecorateDisplay(
		register CatalogBrowserPtr		dbp,
		Boolean							redrawDisplay
	)
{
		Rect							viewRect;
		Fixed							xRatio;
		Fixed							yRatio;
		
		if (redrawDisplay) {
			/*
			 * Hide the panels and erase the window to avoid flashing.
			 */
			(void) SDPHidePanel(DBR.panelHandle);
			(void) SDPHidePanel(DBR.pdPanelHandle);
			EraseRect(&WINDOW->portRect);
		}
		/*
		 * Re-adjusting the display multi-step process, working from the outside
		 * edges towards the center. The reason for this is that the inside
		 * elements are positioned with respect to their surroundings. xRatio and
		 * yRatio reposition the "elastic" elements. The left edge of the local
		 * record name provides the basis for all horizontal repositioning.
		 */
		RECT(kWindowShape) = WINDOW->portRect;
		xRatio = FixRatio(RECT(kWindowShape).right, kMinWindowWidth);
		yRatio = FixRatio(RECT(kWindowShape).bottom, kMinWindowHeight);
		/*
		 * Position the record name box so it fills half of the right side of
		 * the display.
		 */
		SetRect(
			&RECT(kInLocalRecordName),
			HorizontalRatio(kRecordNameLeft),
			OffsetFromTop(kRecordNameTop, kWindowShape),
			OffsetFromRight(-kRecordTypeRightIndent, kWindowShape),
			OffsetFromTop(kRecordNameTop + kRecordNameHeight, kWindowShape)
		);
		RECT(kInLocalRecordName).right =
			RECT(kInLocalRecordName).left
			+ ((width(RECT(kInLocalRecordName)) - kRecordNameHSeparator) / 2);
		/*
		 * Position the record type box with respect to the record name box
		 */
		SetRect(
			&RECT(kInLocalRecordType),
			OffsetFromRight(kRecordNameHSeparator, kInLocalRecordName),
			RECT(kInLocalRecordName).top,
			OffsetFromRight(-kRecordTypeRightIndent, kWindowShape),
			RECT(kInLocalRecordName).bottom
		);
		/*
		 * The spinner size and position are fixed.
		 */
		RECT(kInSpinnerRect).left = OffsetFromLeft(kSpinnerLeft, kWindowShape);
		RECT(kInSpinnerRect).top = OffsetFromTop(kSpinnerTop, kWindowShape);
		SetFixedShape(kInSpinnerRect, kSpinnerSize, kSpinnerSize);
		/*
		 * The lists are all fixed on their left and right edges (and the data
		 * display list is fixed on its left, right, and bottom edges).
		 */
		SetListShape(
			kInAttributeTypeList,
			kInLocalRecordName,
			kTypeListHeight,
			DBR.typeFont.lineHeight
		);
		SetListShape(
			kInAttributeValueList,
			kInAttributeTypeList,
			kValueListHeight,
			DBR.valueFont.lineHeight
		);
		SetListShape(
			kInErrorLog,
			kInAttributeValueList,
			kErrorLogHeight,
			DBR.labelFont.lineHeight
		);
		SetRect(
			&RECT(kInDataDisplayList),
			RECT(kWindowShape).left,
			OffsetFromBottom(kListVSeparator, kInErrorLog),
			OffsetFromRight(-kListRightIndent, kWindowShape),
			OffsetFromBottom(-kScrollBarWidth, kWindowShape)
		);
		AdjustListTop(dbp, kInDataDisplayList, DBR.valueFont.lineHeight);
		/*
		 * The icon display popup is positioned a suitable distance above the data
		 * display list and just to the left of the column of lists.
		 */
		RECT(kInGetCatalogIconPopup).left =
			OffsetFromLeft(
					-(kIconPopupWidth + kIconPopupHSeparator),
					kInLocalRecordName
				);
		RECT(kInGetCatalogIconPopup).top = 
			OffsetFromTop(
					-(kIconPopupHeight
							+ kDirIconVOffset
							+ kDirIconHeight
							+ kListVSeparator
						),
					kInDataDisplayList
				);
		SetFixedShape(kInGetCatalogIconPopup, kIconPopupWidth, kIconPopupHeight);
		/*
		 * The icon displays are fixed with respect to the icon popup
		 */
		RECT(kInGetCatalogIconItself).left =
			OffsetFromLeft(kDirIconHOffset, kInGetCatalogIconPopup);
		RECT(kInGetCatalogIconItself).top =
			OffsetFromBottom(kDirIconVOffset, kInGetCatalogIconPopup);
		SetFixedShape(kInGetCatalogIconItself, kDirIconWidth, kDirIconHeight);
		RECT(kInSDPCatalogIcon).left =
			OffsetFromLeft(-kSDPIconHOffset, kInGetCatalogIconPopup);
		RECT(kInSDPCatalogIcon).top = RECT(kInGetCatalogIconItself).top;
		SetFixedShape(kInSDPCatalogIcon, kSDPIconWidth, kSDPIconHeight);
		/*
		 * Only now can we reposition the SDP Panel itself.
		 */
		SetRect(
			&RECT(kInSDPPanel),
			OffsetFromLeft(kSDPLeft, kWindowShape),
			OffsetFromTop(kBrowserTop + kBrowserHeight + kVSeparator, kWindowShape),
			OffsetFromLeft(-kHSeparator, kInLocalRecordName),
			OffsetFromTop(-kVSeparator, kInGetCatalogIconPopup)
		);
		/*
		 * Position the Browser Popup so it is centered over the SDP Panel
		 */
		SetRect(
			&RECT(kInBrowsePopup),
			OffsetFromLeft(kBrowserPopupInset, kInSDPPanel),
			OffsetFromTop(kBrowserTop, kWindowShape),
			OffsetFromRight(-kBrowserPopupInset, kInSDPPanel),
			OffsetFromTop(kBrowserTop + kBrowserHeight, kWindowShape)
		);
		/*
		 * The Radio Button top and left edges are fixed with respect to the SDP
		 * Panel, their height is fixed and their width is such that they do not
		 * hit the catalog icon.
		 */
		SetRect(
			&RECT(kInShowEntityRadioButton),
			RECT(kInSDPPanel).left,
			OffsetFromBottom(kVSeparator, kInSDPPanel),
			OffsetFromLeft(-kHSeparator, kInSDPCatalogIcon),
			OffsetFromBottom(kVSeparator + kRadioButtonHeight, kInSDPPanel)
		);
		RECT(kInShowCatalogRadioButton) = RECT(kInShowEntityRadioButton);
		OffsetRect(
			&RECT(kInShowCatalogRadioButton),
			0,
			(kRadioButtonHeight + kRadioButtonVSeparator)
		);
		/*
		 * The record length list is fixed with respect to the top
		 * of the data display list.
		 */
		RECT(kInRecordLengthText).left = RECT(kInSDPPanel).left;
		RECT(kInRecordLengthText).top =
			OffsetFromTop(
				-(kRecordLengthHeight + kRecordLengthVSeparator),
				kInDataDisplayList
			);
		SetFixedShape(kInRecordLengthText, kRecordLengthWidth, kRecordLengthHeight);
		/*
		 * That's all, folks.
		 */
		if (redrawDisplay) {
			/*
			 * Now, reposition the SDP Panel, the lists, and controls.
			 */
			RepositionPanel(dbp, DBR.panelHandle);
			RepositionPanel(dbp, DBR.pdPanelHandle);
			RepositionList(DBR.errorLog, &RECT(kInErrorLog));
			RepositionList(DBR.attributeTypeList, &RECT(kInAttributeTypeList));
			RepositionList(DBR.attributeValueList, &RECT(kInAttributeValueList));
			RepositionList(DBR.dataDisplayList, &RECT(kInDataDisplayList));
			RepositionList(DBR.directoryDisplayList, &RECT(kInDataDisplayList));
			if (DBR.setupCatalogList != NULL) {
				viewRect = RECT(kInSDPPanel);
				InsetRect(&viewRect, 4, 4);
				RepositionList(DBR.setupCatalogList, &viewRect);
			}
			RepositionControl(dbp,
				DBR.showEntityRadioControl, kInShowEntityRadioButton);
			RepositionControl(dbp,
				DBR.showDSSpecRadioControl, kInShowCatalogRadioButton);
			RepositionControl(dbp, DBR.browsePopupControl, kInBrowsePopup);
			RepositionControl(dbp, DBR.iconPopupControl, kInGetCatalogIconPopup);
			UpdateBrowserWindow(dbp);
			switch (DBR.browserChoice) {
			case kBrowseSDPPanel:
				(void) SDPShowPanel(DBR.panelHandle);
				break;
			case kBrowsePDPanel:
				(void) SDPShowPanel(DBR.pdPanelHandle);
				break;
			case kBrowseSetupCatalog:
				break;
			}
			ValidRect(&WINDOW->portRect);
		}
}

void
AdjustListBottom(
		register CatalogBrowserPtr		dbp,
		WindowFocus						index,
		short							cellHeight
	)
{
		short							listHeight;
		
		/*
		 * We want our lists to have an integral number of lines of text.
		 */
		listHeight = height(RECT(index));
		listHeight -= (listHeight % cellHeight);
		RECT(index).bottom = RECT(index).top + listHeight;
}
		
void
AdjustListTop(
		register CatalogBrowserPtr		dbp,
		WindowFocus						index,
		short							cellHeight
	)
{
		short							listHeight;
		
		/*
		 * We want our lists to have an integral number of lines of text.
		 */
		listHeight = height(RECT(index));
		listHeight -= (listHeight % cellHeight);
		RECT(index).top = RECT(index).bottom - listHeight;
}

void
RepositionPanel(
		register CatalogBrowserPtr		dbp,
		SDPPanelHandle					panelHandle
	)
{
		OSErr							status;
		
		status = SDPMovePanel(
					panelHandle,
					RECT(kInSDPPanel).left,
					RECT(kInSDPPanel).top
				);
		LogStatus(DBR.errorLog, status, "\pSDPMovePanel");
		status = SDPSizePanel(panelHandle,
					width(RECT(kInSDPPanel)),
					height(RECT(kInSDPPanel))
				);	
		LogStatus(DBR.errorLog, status, "\pSDPSizePanel");
}

/*
 * Note: this does not draw the list content. The Decorator has
 * erased the window, hidden controls, and invalidated everything.
 */
void
RepositionList(
		ListHandle						theList,
		const Rect						*newRect
	)
{
#define LIST	(**theList)

		Boolean							offScreen;
		Rect							viewRect;
		Point							cellSize;

		offScreen = (LIST.rView.left >= 4096);
		viewRect = *newRect;
		if (offScreen)
			OffsetRect(&viewRect, 4096, 4096);
		LIST.rView = viewRect;
		MoveControl(
			LIST.vScroll,
			LIST.rView.right - (kScrollBarWidth - 1),
			LIST.rView.top - 1
		);
		LSize(
			width(viewRect) - kScrollBarWidth,
			height(viewRect),
			theList
		);
		cellSize.v = LIST.cellSize.v;
		cellSize.h = width(LIST.rView);
		LCellSize(cellSize, theList);
#undef LIST
}

/*
 * DecorateControl moves and/or resizes one of the controls. It is called for the
 * popup menus and radio buttons. The control was hidden becomes visible.
 */
void
RepositionControl(
		register CatalogBrowserPtr		dbp,
		ControlHandle					theControl,
		WindowFocus						index
	)
{
		MoveControl(theControl, RECT(index).left, RECT(index).top);
		SizeControl(theControl, width(RECT(index)), height(RECT(index)));
}
