/*									DrawStrings.c								*/
/*
 * DTS Catalog Peek
 * DrawStrings.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * This file contains utility routines for drawing text and managing RStrings.
 */

#include "DTSCatalogPeek.h"
#pragma segment SupportCode

static const short	kCharEllipsis = 0xC9;		/* ... char	*/

/*
 * MakeRStringHandle
 * Store the RString in a handle, return the handle.
 * Failure on memory errors. Invalid RStrings return
 * zero-length handles.
 */
RStringHandle
MakeRStringHandle(
		const RStringPtr				rStringPtr
	)
{
		RStringHandle				result;
		
		result = (RStringHandle)
			NewHandle(RStringLength(rStringPtr));
		FailNIL(result);
		FailOSErr(
			OCECopyRString(
				rStringPtr,
				*result,
				rStringPtr->dataLength
			)
		);
		return (result);
}

/*
 * SetRStringHandle
 * Store the RString in the handle, adjusting its
 * size as needed. Failure on memory errors. Invalid
 * RStrings return zero-length handles.
 */
void
SetRStringHandle(
		RStringHandle					rStringHandle,
		const RStringPtr				rStringPtr
	)
{
		if (rStringPtr == NULL)
			SetHandleSize((Handle) rStringHandle, 0);
		else {
			SetHandleSize(
				(Handle) rStringHandle,
				RStringLength(rStringPtr)
			);
		}
		FailMemError();
		FailOSErr(
			OCECopyRString(
				rStringPtr,
				*rStringHandle,
				rStringPtr->dataLength
			)
		);
}

/*
 * Draw an RString inside of a boundary frame
 */
void
DrawRStringInFrame(
		const RStringHandle				rStringHandle,
		const Rect						*viewRect
	)
{
		Rect						myViewRect;
		
		myViewRect = *viewRect;
		ValidRect(&myViewRect);
		FrameRect(&myViewRect);
		InsetRect(&myViewRect, 2, 2);
		if (IsRStringHandle(rStringHandle))
			DrawRStringHandle(rStringHandle, &myViewRect, normal);
		else {
		 	EraseRect(&myViewRect);
		}
}

/*
 * DrawRStringHandle
 * Draw an RString stored in a handle.
 */
void
DrawRStringHandle(
		const RStringHandle				rStringHandle,
		const Rect						*boundsRect,
		short							textFace
	)
{
		unsigned short				handleState;
		
		if (rStringHandle == NULL
		 || GetHandleSize((Handle) rStringHandle) == 0)
		 	DrawWideString("\p", boundsRect, textFace);
		else {
			handleState = HGetState((Handle) rStringHandle);
			HLock((Handle) rStringHandle);
			DrawRString(*rStringHandle, boundsRect, textFace);
			HSetState((Handle) rStringHandle, handleState);
		}
}

/*
 * DrawRString
 * Draw an RString in the given rectangle. Note: if the
 * source is, originally, in a handle, make sure it is
 * locked before you call this.
 *
 * Note that DrawRString transparently handles multiple
 * scripts.
 */
void
DrawRString(
		const RStringPtr				rStringPtr,
		const Rect						*boundsRect,
		short							textFace
	)
{
		short			oldFont;
		Boolean			mustFixScript;
		
		if (rStringPtr == NULL) {
			DrawWideString(
				"\pnull",
				boundsRect,
				italic
			);
		}
		else if (OCEValidRString(rStringPtr, kOCEGenericInsensitive) == FALSE)
			DrawWideString("\pInvalid", boundsRect, italic);
		else {
			mustFixScript =	(FontScript() != rStringPtr->charSet);
			if (mustFixScript) {
				oldFont = (*qd.thePort).txFont;
				TextFont(
					GetScript(
						rStringPtr->charSet,
						smScriptAppFond
					)
				);
			}
			DrawWideText(
				(Ptr) rStringPtr->body,
				0,
				rStringPtr->dataLength,
				boundsRect,
				textFace
			);
			if (mustFixScript)
				TextFont(oldFont);
		}
}

/*
 * RStringWidth
 * "StringWidth" for RStrings (You've set the font and size and whatever).
 */
unsigned short
RStringWidth(
		const RStringPtr				rStringPtr
	)
{
		short						oldFont;
		Boolean						mustFixScript;
		unsigned short				result;
		
		if (rStringPtr == NULL)
			result = 0;
		else if (OCEValidRString(rStringPtr, kOCEGenericInsensitive) == FALSE)
			result = StringWidth("\pInvalid");
		else {
			mustFixScript =	(FontScript() != rStringPtr->charSet);
			if (mustFixScript) {
				oldFont = (*qd.thePort).txFont;
				TextFont(
					GetScript(
						rStringPtr->charSet,
						smScriptAppFond
					)
				);
			}
			result = TextWidth(
				(Ptr) rStringPtr->body,
				0,
				rStringPtr->dataLength
			);
			if (mustFixScript)
				TextFont(oldFont);
		}
		return (result);
}

/*
 * DrawWideString
 * Draw the string within the alloted space. First try "full
 * width", then "condensed", then shrink the and append '...'
 */
void
DrawWideString(
		const StringPtr					stringPtr,
		const Rect						*boundsRect,
		short							textFace
	)
{
		DrawWideText(
			(Ptr) &stringPtr[1],
			0,
			stringPtr[0],
			boundsRect,
			textFace
		);
}

/*
 * DrawWideText
 * Draw the text within the alloted space. First try "full
 * width", then "condensed", then shrink the and append '...'
 */
void
DrawWideText(
		Ptr							textPtr,
		unsigned short				offset,
		unsigned short				textLength,
		const Rect					*boundsRect,
		short						textFace
	)
{
		short				oldFace;
		Boolean				needEllipsis;
		Ptr					theTextPtr;
		short				fieldWidth;
		FontInfo			fontInfo;

		if (textPtr == NULL) {
			textPtr = (Ptr) "<null>";
			offset = 0;
			textLength = 6;
		}
		if (textLength < 0 || textLength > 256) {
			/*
			 * This is probably a bug.
			 */
			textLength = 128;
		}
		/*
		 * It would be more elegant to erase only the area to
		 * the right of the text, but that * gets rather messy
		 * if we have to deal with ellipses.
		 */
		EraseRect(boundsRect);
		fieldWidth = width(*boundsRect);
		theTextPtr = &textPtr[offset];
		oldFace = (*qd.thePort).txFace;
		TextFace(textFace);
		needEllipsis = FALSE;
		if (TextWidth(theTextPtr, 0, textLength) > fieldWidth) {
			TextFace(textFace | condense);
			if (TextWidth(theTextPtr, 0, textLength) > fieldWidth) {
				fieldWidth -= CharWidth(kCharEllipsis);
				needEllipsis = TRUE;
				while (textLength > 1
				 && TextWidth(theTextPtr, 0, textLength) > fieldWidth)
					--textLength;
			}
		}
		GetFontInfo(&fontInfo);
		EraseRect(boundsRect);
		MoveTo(boundsRect->left, boundsRect->top + fontInfo.ascent);
		DrawText(theTextPtr, 0, textLength);
		if (needEllipsis)
			DrawChar(kCharEllipsis);
		TextFace(oldFace);
}

