/*										Failure.c								*/
/*
 * Failure.c
 * Copyright  1992-93 Apple Computer Inc. All rights reserved.
 * The Failure module is independent of DTS Catalog Peek. It allows application
 * programs (or code resources) to define a coherent error management strategy.
 * The main() function must establish a last-chance failure handler as follows:
 *
 *		InitializeMacintosh();	// All normal initializations
 *		TRY {
 *			InitializeApplication();
 *			ProcessEvents();
 *			ApplicationExit();
 *		}
 *		CATCH {
 *			DisplayFatalErrorAlert(STATUS);
 *			NO_PROPAGATE;		// This is essential
 *		}
 *		ENDTRY;
 *		ExitToShell();
 * ProcessEvents will normally be written as:
 *		while (gQuitNow == FALSE) {
 *			TRY {
 *				ProcessOneEvent();
 *			}
 *			CATCH {
 *				if (IsNonFatalError(STATUS)) {
 *					DisplayNonFatalAlert(STATUS);
 *					NO_PROPAGATE;
 *				}
 *			}
 *			ENDTRY;
 *		}
 * Where DisplayFatalErrorAlert displays the error
 * status and an "Exit" button, while DisplayNonFatalAlert
 * displays the error and a choice of "Exit" or "Continue"
 * Lower-level functions can also handle errors in this
 * fashion.
 *
 * This version of Failure is stripped down. A better
 * implementation would also pass a message index parameter
 * that locates a more extensive explanation of the error
 * in an application-specific STR# resource.
 *
 * This version is also intended for program debugging.
 * All errors store the function name and offset into
 * a global string.
 *
 * Globals
 *	gFailurePtr
 *		This points to the current failure record.
 *		Failure records are created on the stack
 *		by the TRY macro (and are deleted by the
 *		ENDTRY macro). A failure record contains
 *		the setjmp/longjmp context and the current
 *		error status. It could (and should) be extended
 *		to include a message index.
 *	gDebugOnError
 *		This is a programmer convenience. If set
 *		(usually by the debugger), Failure will trap
 *		into the debugger when it is called. It should
 *		never be set in normal circumstances.
 *	gFailureLocation
 *		This string contains the name of the module
 *		that called Failure and the offset of the
 *		call within that module. This uses the same
 *		algorithm used by the MacsBug sc6 command.
 *		Normally, the data would only be displayed
 *		during debugging.
 */
#include "Failure.h"
#pragma segment SupportCode

/*
 * This inline copies the value of A6 (the frame pointer) to D0.
 * It is needed in order to get the caller's function name.
 */
unsigned long *__GetA6(void) = { 0x200E };	/* move.l A6,D0	*/

/*
 * These global values are needed to handle failure. Only gFailurePtr
 */
FailureRecordPtr		gFailurePtr;
Boolean					gDebugOnError;
Str255					gFailureLocation;
#ifndef pstrcpy
/*
 * Cheap 'n dirty pascal string copy routine.
 */
#define pstrcpy(dst, src) do {							\
		StringPtr	_src = (src);						\
		BlockMove(_src, dst, _src[0] + 1);				\
	} while (0)
/*
 * Cheap 'n dirty pascal string concat.
 */
#define pstrcat(dst, src) do {							\
		StringPtr		_dst = (dst);					\
		StringPtr		_src = (src);					\
		short			_len;							\
		_len = 255 - _dst[0];							\
		if (_len > _src[0]) _len = _src[0];				\
		BlockMove(&_src[1], &_dst[1] + _dst[0], _len);	\
		_dst[0] += _len;								\
	} while (0)
#endif
void						AppendHexLeadingZeros(
		StringPtr				result,
		unsigned long			value,
		short					fieldWidth
	);

/*
 * FailureNeverDebug
 * User-callable function to fail without calling debug.
 * Normally, this is done during error recovery.
 */
void
FailureNeverDebug(
		OSErr			errorStatus
	)
{
		__FormatCallerName(__GetA6(), gFailureLocation);
		__FailureNeverDebug(errorStatus);
}

/*
 * __Failure
 * Private function, called by the Failure macro, to
 * initiate error recovery.
 */
void
__Failure(
		OSErr				errorStatus
	)
{
		if (gDebugOnError)
			DebugStr("\pFailure");
		__FormatCallerName(__GetA6(), gFailureLocation);
		__FailureNeverDebug(errorStatus);
}

/*
 * __FailNIL
 * Private function, called by FailNIL when its argument
 * is NULL. This chooses the proper error status and
 * continues the failure process.
 */
void
__FailNIL(void)
{
		OSErr			status;
		
		status = MemError();
		if (status == noErr)
			status = ResError();
		if (status == noErr)
			status = memFullErr;
		__FormatCallerName(__GetA6(), gFailureLocation);
		__FailureNeverDebug(status);
}

/*
 * __FailNeverDebug
 * Private function. This is the actual "failure" function.
 * It stores the status and uses longjmp to jump to the
 * current failure context's CATCH module.
 * Note that errorStatus is forced non-zero: this is
 * an Ansi C restriction.
 */
void
__FailureNeverDebug(
		OSErr			errorStatus
	)
{
		gFailurePtr->status = errorStatus;
		if (gFailurePtr->status == noErr)
			gFailurePtr->status = abortErr;
		longjmp(
			gFailurePtr->jmpBuf, gFailurePtr->status);
}

/*
 * __FormatCallerName
 * Private function: this peeks up the calling chain list to
 * the Failure function caller. It then scans through the
 * function code to locate the start and end of the function.
 * The end of the function is followed by the MacsBug name
 * (maybe). This is based on ShowStackChain.c by Greg Anderson
 * greggor@apple.com
 */
void
__FormatCallerName(
		unsigned long			*a6,
		StringPtr				callerName
	)
{
		unsigned long				returnAddress;
		register unsigned short		*pc;
		short						maxSearch;
		unsigned short				length;
		unsigned long				pcOffset;
#define pcByte		((unsigned char *) pc)

		returnAddress = a6[1];			/* Return address to Failure caller	*/
		/*
		 * First, look forwards for the symbol name
		 */
		pc = (unsigned short *) returnAddress;
		callerName[0] = 0;
		for (maxSearch = 0; maxSearch < 0x4000; maxSearch++, pc++) {
			if (pc[0] == 0x4E56)		/* link								*/
				break;					/* We've gone too far				*/
			if (pc[0] == 0x4E75			/* rts								*/
			  || pc[0] == 0x4ED0		/* jmp (a0)							*/
			  || pc[0] == 0x4E74) {		/* rtd #N							*/
			  	if (pc[0] == 0x4E74)	/* If rtd, skip over displacement	*/
					pc++;
				pc++;					/* Skip over return instruction		*/
			 	if (pc[0] != 0x4E56) {	/* Link instruction means no name	*/
				 	/*
				 	 * pc now points to the word following the RTS.
				 	 * Use the MacsBug naming conventions to determine
				 	 * the function name, if any. See Appendix D in
				 	 * the MacsBug User's Guide.
				 	 */
				 	 length = pcByte[0] & 0x7F;
				 	 if (length >= 0x20 && length <= 0x7F) {
					 	/*
					 	 * Fixed length format: first byte is in the range
					 	 * 0x20 through 0x7F, the high bit may or may not
					 	 * be set.
					 	 */
				 	 	if ((pcByte[1] & 0x80) != 0) {
				 	 		/*
				 	 		 * Pascal 16-byte class name format. The string
				 	 		 * is stored "method" "class" (each takes
				 	 		 * eight bytes) -- MacsBug swaps the order
				 	 		 * and inserts a '.' (Warning, untested.)
				 	 		 */
				 	 		BlockMove(&pcByte[0], &callerName[10], 8);
				 	 		BlockMove(&pcByte[7], &callerName[1], 8);
				 	 		callerName[10] &= ~0x80;
				 	 		callerName[11] &= ~0x80;
				 	 		callerName[9] = '.';
				 	 		callerName[0] = 17;
				 	 	}
				 	 	else {
				 	 		/*
				 	 		 * Pascal 8 byte format.
				 	 		 */
				 	 		BlockMove(&pcByte[0], &callerName[1], 8);
				 	 		callerName[0] = 8;
				 	 	}
					}
					else if (pcByte[0] >= 0x80 && pcByte[0] <= 0x9F) {
						/*
						 * Variable-length string format. If the length
						 * byte is zero after removing the flag bit,
						 * the next byte has the true (Str255) length.
						 */
						if (length == 0) {
							/*
							 * Step over the flag. Note that pc and pcByte
							 * are the same variable (pc is a "short *"
							 * while pcByte is a "char *"). The next
							 * statement adds 1 to pc in a way that
							 * doesn't cause ANSI compilers to lose
							 * their lunches.
							 */
							pc = (unsigned short *) &pcByte[1];
							length = pcByte[0];
						}
						BlockMove(pcByte, callerName, length + 1);
						callerName[0] = length;
					}
					else {
						/*
						 * Something isn't understandable.
						 */
					}
				}
				break;
			}
		}
		/*
		 * If there is no symbol name, don't search for an offset.
		 */
		pcOffset = 0;
		if (callerName[0] != 0) {
			pc = (unsigned short *) returnAddress;
			for (maxSearch = 0; maxSearch < 0x4000; maxSearch++) {
				if (pc[0] == 0x4E56) {				/* link instruction	*/
					pcOffset =
						((unsigned long) returnAddress)
						- ((unsigned long) pc);
					break;
				}
				--pc;
			}
		}
		if (pcOffset == 0 || callerName[0] == 0)
			AppendHexLeadingZeros(callerName, returnAddress, sizeof returnAddress * 2);
		else {
			pstrcat(callerName, "\p+");
			AppendHexLeadingZeros(callerName, pcOffset, sizeof pcOffset * 2);
		}
#undef pcByte
}

