/*								HandleDirEnumSpec.c							*/
/*
 * DTS Catalog Peek
 * HandleDirEnumSpec.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * The DirEnumSpec returned by GetEnumerateParse contains
 * pointers to ephemeral storage. Use these functions to
 * copy the DirEnumSpec into a handle. The handle contents
 * are private to these functions.
 *
 *	MakeHandleFromDirEnumSpec takes a DirEnumSpec and packages
 *		into a private handle-based structure. It returns
 *		NULL on memory errors.
 *	GetDirEnumSpecFromHandle takes a handle as returned
 *		from MakeHandleDirEnumSpec and stores the necessary
 *		data into the argument DirEnumSpec. It ensures that
 *		the handle is locked (the argument DirEnumSpec contains
 *		pointers to the handle contents) and returns a value
 *		that may be passed to HSetState to restore the
 *		handle locked status:
 *			oldState = GetDirEnumSpecFromHandle(myHandle, &myEnumSpec);
 *			DrawRString(...);
 *			HSetState(myHandle, oldState);
 *
 * To create a handle, get the length of the various
 * components, create a handle, and store offsets into
 * the data in place of the RStringPtr.
 */
#include "DTSCatalogPeek.h"
#pragma segment SupportCode

/*
 * MakeHandleFromDirEnumSpec
 * Package the DirEnumSpec into a handle. Return it.
 */
Handle
MakeHandleFromDirEnumSpec(
		const DirEnumSpec				*dirEnumSpecPtr
	)
{
		Size						handleSize;
		Handle						result;
		unsigned long				offset;
#define SRC			(*dirEnumSpecPtr)
#define SRC_LRID	(SRC.u.recordIdentifier)
#define SRC_DNODE	(SRC.u.dNodeIdentifier)
#define DST			(** ((DirEnumSpec **) result))
#define DST_LRID	(DST.u.recordIdentifier)
#define DST_DNODE	(DST.u.dNodeIdentifier)
#define STORE(src, dst) do {									\
		unsigned short	_len_ = RStringLength(src);				\
		BlockMove((src), ((Ptr) (*result)) + offset, _len_);	\
		(dst) = (RStringPtr) offset;							\
		offset += _len_;										\
	} while (0);

		offset = sizeof (DirEnumSpec);
		switch (SRC.enumFlag & ~kEnumInvisibleMask) {
		default:
			result = NULL;
			break;
		case kEnumDistinguishedNameMask:
		case kEnumAliasMask:
		case kEnumPseudonymMask:
			handleSize = sizeof (DirEnumSpec)
					+ RStringLength(SRC_LRID.recordName)
					+ RStringLength(SRC_LRID.recordType);
			result = NewHandle(handleSize);
			if (result != NULL) {
				DST = SRC;				/* Move constant stuff	*/
				STORE(SRC_LRID.recordName, DST_LRID.recordName);
				STORE(SRC_LRID.recordType, DST_LRID.recordType);
			}
			break;
		case kEnumDNodeMask:
			handleSize = sizeof (DirEnumSpec)
					+ RStringLength(SRC_DNODE.name);
			result = NewHandle(handleSize);
			if (result != NULL) {
				DST = SRC;
				STORE(SRC_DNODE.name, DST_DNODE.name);
			}
			break;
		}
		return (result);
#undef STORE
#undef SRC 
#undef SRC_LRID
#undef SRC_DNODE
#undef DST
#undef DST_LRID
#undef DST_DNODE
}

/*
 * GetDirEnumSpecFromHandle
 * Lock the handle and create a DirEnumSpec whose RStringPtr's
 * point back to the handle string data. Return the handle
 * lock state.
 */
unsigned short
GetDirEnumSpecFromHandle(
		Handle							handleDirEnumSpec,
		DirEnumSpec						*dstEnumSpecPtr
	)
{
		DirEnumSpec						*srcEnumSpecPtr;
		unsigned short					handleState;
#define SRC			(*srcEnumSpecPtr)
#define SRC_LRID	(SRC.u.recordIdentifier)
#define SRC_DNODE	(SRC.u.dNodeIdentifier)
#define DST			(*dstEnumSpecPtr)
#define DST_LRID	(DST.u.recordIdentifier)
#define DST_DNODE	(DST.u.dNodeIdentifier)
#define ADDRESS(i)	\
	((RStringPtr) (((Ptr) srcEnumSpecPtr) + ((unsigned long) (i))))

		handleState = HGetState(handleDirEnumSpec);
		HLock(handleDirEnumSpec);
		srcEnumSpecPtr = (DirEnumSpec *) (*handleDirEnumSpec);
		DST = SRC;				/* Copy fixed info	*/
		switch (SRC.enumFlag & ~kEnumInvisibleMask) {
		default:
			/* Can't happen */
			break;
		case kEnumDistinguishedNameMask:
		case kEnumAliasMask:
		case kEnumPseudonymMask:
			DST_LRID.recordName = ADDRESS(SRC_LRID.recordName);
			DST_LRID.recordType = ADDRESS(SRC_LRID.recordType);
			break;
		case kEnumDNodeMask:
			DST_DNODE.name = ADDRESS(SRC_DNODE.name);
			break;
		}
		return (handleState);
#undef SRC
#undef SRC_LRID
#undef SRC_DNODE
#undef DST
#undef DST_LRID
#undef DST_DNODE
}

