/*										Help.c								*/
/*
 *								H E L P
 */
#ifdef DOCUMENTATION

Title	help

Usage
		DoHelp(
			short		dialogID,			/* DLOG/DITL to show help	*/
			short		topicID				/* STR# with help topics	*/
		)

Description

		Using the specified 'DLOG' and 'STR#' resources, display the
		help dialog and manage user interaction.
		
		The 'STR#' and its associated 'TEXT' and 'styl' resources are
		constructed by the MakeHelp application: see it for a
		description of the help source.
		
		Each 'STR#' item contains the name of a topic.  The associated
		'TEXT' (and 'styl') resource begin at a 'TEXT' resource id as
		specified in the database.
				
		A 'TEXT' resource may also have an associated 'styl' (help style)
		resource that gives optional style runs for the item of the
		same number.  The style resource has the format used for style
		scraps.
		
The Help Resource Header

		The first few strings in the topic STR# contain the following
		data:
			1	The first 'TEXT'/'styl' resource for this help section.
			2	The primary font name (default is geneva).
			3	The monospaced font name (default is monaco)
			4	The font size (default is 9)
			
The Help Dialog

		Here is a sample DLOG/DITL for the help function.

		resource 'DLOG' (DLOG_Help) {
			{  0,  0, 284, 472 },
			dBoxProc, invisible, noGoAway, 0x0, DLOG_Help,
			""
		};
		resource 'DITL' (DLOG_Help) {
			{
				{ 243, 200, 262, 272 },	Button
										{ enabled,	"OK"		},
				{ 243,  80, 262, 160 },	Button
										{ enabled,	"Prev"		},
				{ 243, 312, 262, 392 },	Button
										{ enabled,	"Next"		},
				{  10, 172, 239, 465 },	UserItem
										{ enabled				},
				{  10,   7, 239, 165 },	UserItem
										{ enabled				}
			}
		};
		
Author

		Help is closely based on an article published in The Essential
		MacTutor, Vol 3, written by:
			William Rausch
			Battele Pacific Northwest Labs.
			Richland, WA
		This significantly revised version is by:
			Martin Minow
			minow@apple.com
			
Limitations

		Help requires a version of the Macintosh system that
		supports "Styled TextEdit".
		
		Help compiles under Think C v5.0 and MPW 3.2.
		
Copyright

		This program is Copyright  1989-92 Martin Minow.
		All rights reserved.
		Permission is hereby granted to incorporate this software in
		your applications without fee or specific permission. You may
		copy and distribute the source form of this software provided
		that this copyright notice, this permission notice, and its
		date remain intact and notice is given that copying is by
		permission of Martin Minow and copies are not made or
		distributed for direct commercial advantage and that you
		accept all responsibility for the correct operation of this
		software.  To copy otherwise, or to republish, requires a fee
		and/or specific permission.		
		
#endif

/*
 * This function makes use of four resources, a STR# containing the
 * list of topics, a group of TEXT/styl resources containing the
 * appropriate text and a DLOG/DITL.
 *
 * The STR# nnn resource contains a list of help topics. (nnn is a
 *			parameter to DoHelp()).
 * The TEXT nnn resources contain text for each help topic.
 */

#ifndef THINK_C			/* MPW needs header file #includes				*/
#include <Processes.h>
#include <Errors.h>
#include <Finder.h>
#include <Folders.h>
#include <Script.h>
#include <Types.h>
#include <Resources.h>
#include <QuickDraw.h>
#include <Fonts.h>
#include <Events.h>
#include <Windows.h>
#include <Menus.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Desk.h>
#include <ToolUtils.h>
#include <Memory.h>
#include <Files.h>
#include <OSUtils.h>
#include <OSEvents.h>
#include <DiskInit.h>
#include <Packages.h>
#include <Traps.h>
#include <Lists.h>
#endif
#pragma segment SupportCode

#if 0
#define DEBUG(str)	DebugStr(str)
#else
#define DEBUG(str)	/* Nothing */
#endif

/*
 * Help is intentionally self-contained, even though this results in
 * a few duplicated functions.
 */
#ifndef FALSE
#define	FALSE	0
#define	TRUE	1
#endif
#define height(rect)	((rect).bottom - (rect).top)
#define width(rect)		((rect).right - (rect).left)

/*
 * Various constants
 */
enum {
	sBarWidth			= 15,		/* Scroll bar width					*/
	kActiveButton		= 0,		/* Normal button hilite				*/
	kDisabledButton		= 255		/* Disabled button hilite			*/
};

/*
 * Keycode and keyboard number values.
 */
enum {
	keyEnter			= 0x03,
	keyReturn			= 0x0D,
	keyESC				= 0x1B,
	keyLeftArrow		= 0x1C,
	keyRightArrow		= 0x1D,
	keyUpArrow			= 0x1E,
	keyDownArrow		= 0x1F,
	keyPageUp			= 0x0B,
	keyPageDown			= 0x0C,
	keyHome				= 0x01,
	keyEnd				= 0x04
};

/*
 * This information is shared with MakeHelp.
 */
enum {
	kPlainFont = geneva,			/* ~P font number					*/
	kMonospaceFont = monaco			/* ~M font number					*/
};

enum {								/* The first STR# resources			*/
	kHelp_ResourceBase = 1,			/* first TEXT/styl resource id		*/
	kHelp_MainFont,					/* ~P font name "Palatino"			*/
	kHelp_MonospaceFont,			/* ~M font name "Courier"			*/
	kHelp_FontSize,					/* Point size for ~P				*/
	kHelp_FirstTopic				/* Offset in STR# to first topic	*/
};

/*
 * This defines the DITL.
 */
enum {
	kHelp_OK = 1,				/* The OK (exit) button					*/
	kHelp_Prev,
	kHelp_Next,
	kHelp_Text,
	kHelp_List
};

/*
 * The HelpRecord structure contains our "globals".
 */
typedef struct {
	DialogRecord	theDialog;		/* The dialog itself.				*/
	Rect			textRect;		/* True shape of the help text		*/
	TEHandle		textHandle;		/* -> current help text				*/
	ListHandle		topicList;		/* Help index list					*/
	ControlHandle	prevButton;		/* Prev topic button control		*/
	ControlHandle	nextButton;		/* Next topic button control		*/
	ControlHandle	textScroll;		/* Scroll bar for help text			*/
	short			resourceBase;	/* First TEXT resource				*/
	short			primaryFont;	/* i.e. Geneva						*/
	short			monospaceFont;	/* i.e. Monoco						*/
	short			fontSize;		/* Font size from help database		*/
	short			lineHeight;		/* Text line height					*/
	short			offset;			/* Vert scroll offset				*/
	short			viewLines;		/* Help text viewRect height		*/
	short			maxTopic;		/* Actual number of STR# items		*/
	short			currentTopic;	/* Currently-visible topic			*/
} HelpRecord, *HelpPtr;

/*
 * All functions have a local variable that points to the help record.
 * We access the help record through the HH macro.  A few popular
 * help record items have their own macro shortcut references.
 */
#define	HELP	(*helpPtr) 			/* Always use HELP.xxx				*/
#define DIALOG	((DialogPtr) &HELP.theDialog) /* DialogPtr				*/
#define TEXT	(**HELP.textHandle)	/* Dereferenced TEHandle			*/
#define LIST	(**HELP.topicList)	/* Dereferenced ListHandle		*/

/*
 * Global function
 */
void
DoHelp(
		short				dialogID,	/* -> DLOG/DITL				*/
		short				topicID		/* -> STR# with topics		*/
	);

/*
 * Local functions
 */
static Boolean			SetupHelp(HelpPtr, short, short);
static void				ReadHelp(HelpPtr, short);
static void				ShowHelp(HelpPtr);
static pascal Boolean	HelpFilter(WindowPtr, EventRecord *, short *);
static pascal void		HelpAction(ControlHandle, short);
static void				DisplayHelp(HelpPtr, short);
static void				ScrollHelp(HelpPtr);


/*
 * Mainline help processor.
 */
void
DoHelp(
		short				dialogID,	/* -> DLOG/DITL				*/
		short				topicID		/* -> STR# with topics		*/
	)
{
		GrafPtr				saveGrafport;
		HelpRecord			helpRecord;
		short				i;
		
		GetPort(&saveGrafport);
		/*
		 * Make sure the help record is cleaned out. The exit process
		 * looks at aDefItem, which is set to 1 if the dialog opened.
		 */
		for (i = 0; i <= sizeof helpRecord; i++)
			((char *) &helpRecord)[i] = 0;
		if (SetupHelp(&helpRecord, dialogID, topicID) == FALSE)
			SysBeep(10);
		else {
			ReadHelp(&helpRecord, topicID);	/* Read the help topics	*/
			ShowHelp(&helpRecord);			/* Display the stuff	*/
		}
		if (helpRecord.textHandle != NULL)
			TEDispose(helpRecord.textHandle);
		if (helpRecord.textScroll != NULL)
			DisposeControl(helpRecord.textScroll);
		if (helpRecord.topicList != NULL)
			LDispose(helpRecord.topicList);
		if (helpRecord.theDialog.aDefItem != 0)
			CloseDialog((DialogPtr) &helpRecord.theDialog);
		SetPort(saveGrafport);
}

/*
 * Initialize the help process. Return TRUE if initialized,
 * FALSE on error.
 */
static Boolean
SetupHelp(
		register HelpPtr	helpPtr,
		short				dialogID,
		short				topicID
	)
{
		Handle				handle;
		short				kind;
		Rect				listRect;
		Rect				dataBounds;
		Rect				scrollRect;		/* Text scroll bar			*/
		Rect				dispRect;		/* TextEdit's dispRect		*/
		Rect				viewRect;		/* TextEdit's viewRect		*/
		Point				cell;
		FontInfo			info;
		Str255				work;
		long				tempLong;
		
		if (GetNewDialog(dialogID, DIALOG, (WindowPtr) -1L) == NULL) 
			return (FALSE);
		/*
		 * Read the STR# and get the number of topics.
		 * The number of things in the STR# is in the first two bytes.
		 * Then get the header information.
		 */
		if ((handle = GetResource('STR#', topicID)) == NULL) {
			DEBUG("\pNo Help STR# resource");
			goto failure;
		}
		HELP.maxTopic = *((short *) (*handle));
		ReleaseResource(handle);
		GetIndString(work, topicID, kHelp_ResourceBase);
		StringToNum(work, &tempLong);
		HELP.resourceBase = tempLong;
		GetIndString(work, topicID, kHelp_MainFont);
		GetFNum(work, &HELP.primaryFont);
		GetIndString(work, topicID, kHelp_MonospaceFont);
		GetFNum(work, &HELP.monospaceFont);
		GetIndString(work, topicID, kHelp_FontSize);
		StringToNum(work, &tempLong);
		HELP.fontSize = tempLong;
		/*
		 * We mess around with ports and fonts in order to convince
		 * the Macintosh machinery to display our list data in our font.
		 */
		SetPort(DIALOG);
		TextFont(HELP.primaryFont);
		TextSize(HELP.fontSize);
		MoveWindow(						/* Center the window			*/
			DIALOG,
			(width(qd.screenBits.bounds) - width(DIALOG->portRect)) / 2,
			((height(qd.screenBits.bounds)
			  - GetMBarHeight()
			  - height(DIALOG->portRect)
			 ) / 2) + GetMBarHeight(),
			TRUE
		);
		/*
		 * Get the button control handles for the dialog filter.
		 */
		GetDItem(DIALOG, kHelp_Prev,
				&kind, (Handle *) &HELP.prevButton, &listRect);
		GetDItem(DIALOG, kHelp_Next,
				&kind, (Handle *) &HELP.nextButton, &listRect);
		GetDItem(DIALOG, kHelp_Text,
				&kind, &handle, &HELP.textRect);
		GetDItem(DIALOG, kHelp_List,
				&kind, &handle, &listRect);
		/*
		 * Make room for a frame and a vertical ScrollHelp bar:
		 * listRect		the visible shape of our one-column list.
		 * dataBounds	the number of rows and columns of the list
		 * cell			the screen shape of each list cell
		 */
		InsetRect(&listRect, 1, 1);
		listRect.right -= sBarWidth;
		SetRect(&dataBounds, 0, 0, 1,
				HELP.maxTopic - kHelp_FirstTopic + 1);
		GetFontInfo(&info);
		/*
		 * Fiddle with the index display rectangle to make sure that an
		 * integral number of cells are visible.  This is needed for the
		 * "visible" calculation when processing the <next> or <previous>
		 * buttons. Note that using HELP.lineHeight makes variable-height
		 * lines difficult to manage.
		 */
		HELP.lineHeight = info.ascent + info.descent + info.leading;
		SetPt(&cell, width(listRect), HELP.lineHeight);
		HELP.viewLines = height(listRect) / HELP.lineHeight;
		listRect.bottom = listRect.top + (HELP.viewLines * HELP.lineHeight);
		/*
		 * Fiddle the text box bottom so it's even with the topic box.
		 * the +2 undoes the InsetRect above. We probably should install
		 * a list definition procedure so we can draw the index in
		 * a compressed typeface if necessary.
		 */
		HELP.textRect.bottom = HELP.textRect.top + height(listRect) + 2;
		HELP.topicList = LNew(
			&listRect,					/* Display rectangle			*/
			&dataBounds,				/* List rows and columns		*/
			cell,						/* Size of each cell 			*/
			0,							/* No list definition procedure	*/
			DIALOG,						/* Window that gets the list	*/
			FALSE,						/* Don't draw it yet			*/
			FALSE,						/* No Grow box					*/
			FALSE,						/* No horizontal ScrollBar		*/
			TRUE						/* Vertical ScrollBar			*/
		);
		if (HELP.topicList == NULL)
			goto failure;
		LIST.selFlags = lOnlyOne;
		/*
		 * Build the scroll bar.
		 */
		scrollRect = HELP.textRect;		/* Scroll bar shape				*/
		scrollRect.left = scrollRect.right - sBarWidth;
		HELP.textScroll = NewControl(	/* Text box scroll bar			*/
			DIALOG,						/* Its window					*/
			&scrollRect,				/* Location in window			*/
			"\p",						/* No title						*/
			TRUE,						/* Visible						*/
			0, 0, 0,					/* Initial value, min, max		*/
			scrollBarProc,				/* Control definition id		*/
			(long) helpPtr				/* Linkage to help record		*/
		);
		if (HELP.textScroll == NULL)
			goto failure;
		/*
		 * Build the text edit record.
		 */
		dispRect = HELP.textRect;		/* TextEdit's display rect		*/
		dispRect.right -= sBarWidth;	/* Remove scroll area			*/
		InsetRect(&dispRect, 3, 1);		/* Leave room for frame			*/
		viewRect = dispRect;			/* TextEdit's view rect			*/
		dispRect.bottom = 30000;		/* Very long text				*/
		HELP.textHandle = TEStylNew(&dispRect, &viewRect);
		if (HELP.textHandle == NULL)
			goto failure;
		TEXT.lineHeight = HELP.lineHeight;
		TEXT.fontAscent = info.ascent;
		return (TRUE);					/* Success						*/
failure:
		return (FALSE);					/* Sorry, mainline cleans up	*/
}

/*
 * Read the topics from the STR#, building the help list
 * and the help text pointer in HELP.id[]
 */
static void
ReadHelp(
		register HelpPtr	helpPtr,
		short				topicID
	)
{
		register short		index;		/* Count the topics				*/
		Str255				topicName;	/* Topic name					*/
		Point				cell;		/* List cell					*/
		
		/*
		 * For each entry in the STR#, read the item from the resource
		 * and store it in the cell list.
		 */
		for (index = kHelp_FirstTopic; index <= HELP.maxTopic; index++) {
			GetIndString(topicName, topicID, index);
			SetPt(&cell, 0, index - kHelp_FirstTopic);
			LSetCell(
				&topicName[1],			/* Item to store				*/
				topicName[0],			/* Length of the item			*/
				cell,					/* Item index					*/
				HELP.topicList			/* Into this list				*/
			);
		}
		SetPt(&cell, 0, 0);				/* Select first element and		*/
		LSetSelect(TRUE, cell, HELP.topicList);
		LDoDraw(TRUE, HELP.topicList);	/* Draw the list				*/
}

/*
 * Display the text for the first item in the list, then call
 * ModalDialog with a filter procedure to do all the work.
 * Loop until the user clicks OK.
 */
static void
ShowHelp(
		register HelpPtr	helpPtr
	)
{
		short			itemHit;		/* Returned by modal dialog		*/
		short			topic;			/* New list selection			*/
		Cell			theCell;		/* For list selection			*/
		
		HiliteControl(HELP.textScroll, kDisabledButton);
		/*
		 * Setup the selection and fiddle the viewRect so
		 * it has an integral number of lines of text.
		 */
		HELP.currentTopic = 0;			/* First topic is selected		*/
		HELP.offset = 0;				/* At the top-left corner		*/
		DisplayHelp(helpPtr, 0);		/* Display the first item		*/
		ShowWindow(DIALOG);
		do {
			ModalDialog(HelpFilter, &itemHit);
			switch (itemHit) {
			case kHelp_Prev:			/* Previous topic				*/
				topic = HELP.currentTopic - 1;
				goto new_topic;
			case kHelp_Next:			/* Next topic					*/
				topic = HELP.currentTopic + 1;
new_topic:		if (topic >= 0 && topic <= HELP.maxTopic) {
					/*
					 * Previous/next button click:  un-hilite the
					 * current topic, select the new topic, and
					 * scroll so it's visible.
					 */
					SetPt(&theCell, 0, HELP.currentTopic);
					LSetSelect(FALSE, theCell, HELP.topicList);
					SetPt(&theCell, 0, topic);
					LSetSelect(TRUE, theCell, HELP.topicList);
					/*
					 * This is a less jerkey version of LAutoScroll
					 * -- it scrolls the list in the desired direction.
					 * Since we only get here by a button push, this
					 * should scroll once only.  Unless, that is, you
					 * scroll the topic list offscreen and hit <Prev>
					 * or <Next>. This should be redone so that the
					 * newly selected item moves to the closest edge
					 * of the window. Instead, we scroll once (possibly
					 * in the wrong direction, then use the big hammer
					 * of LAutoScroll to force the selected cell onto
					 * the screen. If it was offscreen, it will move
					 * to the topmost position, which doesn't look
					 * great, but is probably close enough.
					 */
					if (PtInRect(theCell, &LIST.visible) == FALSE) {
						LScroll(
							0, topic - HELP.currentTopic, HELP.topicList
						);
					}
					LAutoScroll(HELP.topicList);
					DisplayHelp(helpPtr, topic);
				}
				break;
			}
		} while (itemHit != kHelp_OK);
}

/*
 * HelpFilter handles activating and updating the scroller and box
 * area.  We must handle the enter/return key and mouse hits in the
 * scroller and text box, passing back TRUE, while passing FALSE back
 * for everything  else so the dialog manager does its thing correctly.
 */
static pascal Boolean
HelpFilter(
		register WindowPtr			dialog,
		EventRecord					*event,
		short						*itemHit
	)
{
		short				part;
		ControlHandle		theControl;
		Point				point;
		Cell				cell;
		short				type;
		Handle				handle;
		Rect				box;
#define helpPtr ((HelpPtr) dialog)

		switch(event->what) {
		case updateEvt:
			if (event->message == (long) dialog) {
				/*
				 * Redraw our window.
				 */
				box = LIST.rView;
				InsetRect(&box, -1, -1);
				EraseRect(&box);
				FrameRect(&box);
				EraseRect(&HELP.textRect);
				FrameRect(&HELP.textRect);
				TEUpdate(&HELP.textRect, HELP.textHandle);
				Draw1Control(HELP.textScroll);
				LUpdate(dialog->visRgn, HELP.topicList);
				/*
				 * Draw a frame around the OK button.
				 */
				GetDItem(dialog, kHelp_OK, &type, &handle, &box);
				InsetRect(&box, -4, -4);
				PenSize(3, 3);
				FrameRoundRect(&box, 16, 16);
				PenNormal();
			}
			break;						/* Modal Dialog does the rest	*/

		case keyDown:
		case autoKey:
			switch (event->message & charCodeMask) {
			case keyEnter:
			case keyReturn:
			case keyESC:		*itemHit = kHelp_OK;	return (TRUE);
			case keyLeftArrow:	*itemHit = kHelp_Prev;	return (TRUE);
			case keyRightArrow:	*itemHit = kHelp_Next;	return (TRUE);
			case keyUpArrow:	part = inUpButton;		goto FakeButton;
			case keyDownArrow:	part = inDownButton;	goto FakeButton;
			case keyPageUp:		part = inPageUp;		goto FakeButton;	
			case keyPageDown:	part = inPageDown;
FakeButton:		HelpAction(HELP.textScroll, part);
				return (TRUE);
			case keyHome:
				part = GetCtlMin(HELP.textScroll);
				goto FakeScroll;
			case keyEnd:
				part = GetCtlMax(HELP.textScroll);
FakeScroll:		SetCtlValue(HELP.textScroll, part);
				ScrollHelp(helpPtr);
				return (TRUE);
			}
			break;
						
		case mouseDown:
			point = event->where;		/* Global mouse point			*/
			GlobalToLocal(&point);		/* Mouse point in window		*/
			part = FindControl(point, dialog, &theControl);
			if (part > 0 && theControl == HELP.textScroll) {
				/*
				 * Click in the text box scroll bar.
				 * Do the normal TextEdit scroll stuff.
				 */
				if (part == inThumb) {
					if (TrackControl(theControl, point, NULL))
						ScrollHelp(helpPtr);
				}
				else {
					TrackControl(theControl, point, (ProcPtr) HelpAction);
				}
			}
			else if (PtInRect(point, &LIST.rView)
				  || theControl == LIST.vScroll) {
				/*
				 * Click in the topics list: find out which topic was
				 * selected and display the proper help resource.
				 */
				LClick(point, event->modifiers, HELP.topicList);
				SetPt(&cell, 0, 0);
				if (LGetSelect(TRUE, &cell, HELP.topicList)) {
					if (cell.v != HELP.currentTopic)
						DisplayHelp(helpPtr, cell.v);
				}
			}
			break;			
		}
		return (FALSE);
#undef helpPtr
}

/*
 * HelpAction is called by the Toolbox while executing the TrackControl
 * routine.  It has to take care of scrolling the text when the user
 * clicks on the up/down arrow or page parts of the scroll bar.
 * HELP.viewLInes is the number of lines per screen. We subtract 1 so
 * there is some extra context.
 */
static pascal void
HelpAction(
		register ControlHandle	theControl,
		short					partcode
	)
{
		register HelpPtr	helpPtr;
		short				value;
		short				delta;
		
		helpPtr = (HelpPtr) (*theControl)->contrlRfCon;
		switch (partcode) {
		case inUpButton:	delta = -1;						break;
		case inPageUp:		delta = -(HELP.viewLines - 1);	break;		
		case inDownButton:	delta =  1;						break;
		case inPageDown:	delta =  (HELP.viewLines - 1);	break;
		default:			return;		/* Mouse exited from control	*/
		}
		value = GetCtlValue(theControl) + delta;
		SetCtlValue(theControl, value);
		ScrollHelp(helpPtr);
}

/*
 * DisplayHelp is a fairly general "display a TEXT resource in a TextEdit
 * record" that loads the TEXT specified by the topic parameter and
 * sets up the scroll bar as needed. It handles styled text.
 */
static void
DisplayHelp(
		register HelpPtr	helpPtr,	/* The help record				*/
		short				topic		/* The selected list cell		*/
	)
{
		Handle				theHelpText;
		long				textHeight;
		StScrpHandle		styleHandle;
		TextStyle			oldStyle;
		TextStyle			newStyle;
		
		SetCursor(*GetCursor(watchCursor));		/* Styled text is slow	*/
	 	TEDeactivate(HELP.textHandle);
	 	TESetSelect(0, 32767, HELP.textHandle);	
	 	TEDelete(HELP.textHandle);
	 	/*
	 	 * Force the viewRect offscreen so the style record doesn't
	 	 * cause the text to be shown twice.
	 	 */
		OffsetRect(&TEXT.viewRect, 4096, 4096);	/* Don't draw now	*/
		/*
		 * Reset scroll bar, reposition the help text, and read and
		 * draw the text.  If we draw the text before moving the
		 * scroll bar it jerks around if the scroll point moves.
		 */
		SetCtlValue(HELP.textScroll, 0);
		ScrollHelp(helpPtr);
		/*
		 * Read the text resource and draw the text.
		 */
 		theHelpText = GetResource('TEXT', topic + HELP.resourceBase);
 		if (theHelpText != NULL) {
 			styleHandle = (StScrpHandle)
 				GetResource('styl', topic + HELP.resourceBase);
 			HLock(theHelpText);
	 		TESetSelect(0, 32767, HELP.textHandle);	
	 		TEStylInsert(
	 			*theHelpText,
	 			SizeResource(theHelpText),
	 			styleHandle,			/* OK if it's NULL				*/
	 			HELP.textHandle
	 		);
			HUnlock(theHelpText);
			ReleaseResource(theHelpText);
			if (styleHandle != NULL) {
				ReleaseResource((Handle) styleHandle);
				/*
				 * Replace the internal fonts (geneva and monaco)
				 * by the user's specific fonts.  If the font
				 * is missing, it will end up as systemFont.
				 */
				if (HELP.primaryFont != kPlainFont) {
					TESetSelect(0, 32767, HELP.textHandle);
					oldStyle.tsFont = kPlainFont;
					newStyle.tsFont = HELP.primaryFont;
					TEReplaceStyle(
						doFont, &oldStyle, &newStyle,
						FALSE, HELP.textHandle
					);
				}
				if (HELP.monospaceFont != kMonospaceFont) {
					TESetSelect(0, 32767, HELP.textHandle);
					oldStyle.tsFont = kMonospaceFont;
					newStyle.tsFont = HELP.monospaceFont;
					TEReplaceStyle(
						doFont, &oldStyle, &newStyle,
						FALSE, HELP.textHandle
					);
				}
			}
		}
 		else {
 			/*
 			 * The data couldn't be read from the resource.
 			 * This "can't happen," so we just put up a
 			 * blank <nnn> message.
 			 */
 			unsigned char		foobar[16];

			NumToString((long) topic + HELP.resourceBase, foobar);
			TEInsert(&foobar[1], (long) foobar[0], HELP.textHandle);
		}
		OffsetRect(&TEXT.viewRect, -4096, -4096);	/* Time to draw it	*/
		/*
		 * Hack: force text edit redraw.  If we just InvalRect the
		 * text edit rectangle, the update event redraws everything
		 * so we do the whole update thing here.
		 */
		InvalRect(&TEXT.viewRect);
		BeginUpdate(DIALOG);
		EraseRect(&TEXT.viewRect);
		TEUpdate(&DIALOG->portRect, HELP.textHandle);
		EndUpdate(DIALOG);
		/*
		 * Still need the scroll bar?  Note that TEGetHeight is
		 * a long, but the control max is an integer.  This will
		 * cause problems if we have more than two thousand lines
		 * of text in a help record, so we use the number of
		 * lines rather than the number of pixels.
		 */
		textHeight =
			TEGetHeight(TEXT.nLines, 0, HELP.textHandle)
			/ HELP.lineHeight;
		if (textHeight <= HELP.viewLines)
			HiliteControl(HELP.textScroll, kDisabledButton);
		else {
			HiliteControl(HELP.textScroll, kActiveButton);
			SetCtlMax(HELP.textScroll, textHeight - HELP.viewLines);
		}
	 	HELP.currentTopic = topic;			/* Remember the topic	*/
	 	HiliteControl(HELP.prevButton,
	 		(topic <= 0) ? kDisabledButton : kActiveButton);
	 	HiliteControl(HELP.nextButton,
	 		(topic >= (HELP.maxTopic - kHelp_FirstTopic))
	 			? kDisabledButton : kActiveButton);
		InitCursor();
}

/*
 * Scroll the help information.
 */
static void
ScrollHelp(helpPtr)
register HelpPtr	helpPtr;
{
		short		delta;
		
		delta = HELP.offset - GetCtlValue(HELP.textScroll);
		TEScroll(0, delta * HELP.lineHeight, HELP.textHandle);
		HELP.offset -= delta;
}

