/*								ListManagement.c							*/
/*
 * DTS Catalog Peek
 * ListManagement.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * Various, fairly generic List Manager routines.
 *
 * DisposeList is here, too.
 */
#include "DTSCatalogPeek.h"
#include <Palettes.h>
#define IS_COLOR(port) (((((CGrafPtr) (port))->portVersion) & 0xC000) != 0)
#pragma segment SupportCode

void								MoveList(
		ListHandle						listHandle,
		short							leftEdge,
		short							topEdge
	);

/*
 * This code sequence is copied into the ListProc handle. It
 * is designed so we don't have to flush the instruction
 * and data caches.
 */
static const short gDummyLDEF[] = {
		0x207A,			/*		movea.l	procPtr,a0			*/
		0x0004,			/*				<offset to procPtr>	*/
		0x4ED0,			/*		jmp		a0 					*/
		0x0000, 0x0000	/*		dc.l	<Drawing Proc here>	*/
};

/*
 * CreateList
 * Create a list using the specified font characteristics. If the caller
 * designates a drawing defProc, initialize that, too.
 */
ListHandle
CreateList(
		register CatalogBrowserPtr		dbp,
		const Rect						*viewRect,
		short							listFontNumber,
		short							listFontSize,
		ProcPtr							listDefProc
	)
{
		ListHandle					resultList;
		short						listHeight;
		Point						cellSize;
		Rect						dataBounds;
		Rect						listRect;
		Handle						drawProcHdl;
		FontInfo					info;
		short						listFontHeight;
		
		TextFont(listFontNumber);
		TextSize(listFontSize);
		GetFontInfo(&info);
		/*
		 * Define the list drawing area, adjusting the list area height so an
		 * integral number of lines will be drawn. Note: there must be enough
		 * room around the viewRect to draw a focus rectangle.
		 */
		listRect = *viewRect;
		listRect.right -= kScrollBarWidth;
		listFontHeight = info.ascent + info.descent + info.leading;
		listHeight = height(listRect);
		listHeight -= (listHeight % listFontHeight);
		listRect.bottom = listRect.top + listHeight;
		SetPt(&cellSize, width(listRect), listFontHeight);
		/*
		 * Define a one-column list.
		 */
		SetRect(&dataBounds, 0, 0, 1, 0);
		resultList = LNew(
				&listRect,				/* Viewing area		*/
				&dataBounds,			/* Rows and col's	*/
				cellSize,				/* Element size		*/
				0,						/* No defProc yet	*/
				WINDOW,					/* Display window	*/
				TRUE,					/* Draw it			*/
				FALSE,					/* No grow box		*/
				FALSE,					/* No horiz scroll	*/
				TRUE					/* Vertical scroll	*/
			);
		FailNIL(resultList);
		FailMemError();
		(**resultList).selFlags = lOnlyOne;
		(**resultList).refCon = (long) dbp;
		if (listDefProc != NULL) {
			/*
			 * Create a handle with a callback to our
			 * private drawing function.
			 */
			FailOSErr(
				PtrToHand(
					gDummyLDEF,
					&drawProcHdl,
					sizeof gDummyLDEF
				)
			);
			BlockMove(
				(Ptr) &listDefProc,
				&((short *) *drawProcHdl)[3],
				sizeof listDefProc
			);
			(**resultList).listDefProc = drawProcHdl;
		}
		/*
		 * Draw the list frame on creation.
		 */
		UpdateList(resultList, FALSE);
		return (resultList);
}

/*
 * UpdateList
 * Update a single list by framing it, updating it, and
 * redrawing the focus rectangle, if necessary.
 */
void
UpdateList(
		ListHandle						listHandle,
		Boolean							hasFocus
	)
{
		Rect							viewRect;
		PenState						savePenState;
		
		GetPenState(&savePenState);
		PenNormal();
		viewRect = (**listHandle).rView;
		InsetRect(&viewRect, -1, -1);
		FrameRect(&viewRect);
		if (hasFocus)
			DrawFocusRect(listHandle, patOr);
		LUpdate((**listHandle).port->visRgn, listHandle);
		SetPenState(&savePenState);
}

/*
 * DrawFocusRect
 * Draw a focus rectangle around a specified list. If penMode is patOr, this
 * is the selected list. If it is patBic, this is the deselected list, and the
 * focus rectangle is erased. Well, that's the idea. However when the list
 * extends to the edge of the window, the focus rectangle can't be drawn
 * so it looks reasonable. Thus, this code does nothing.
 */
void
DrawFocusRect(
		ListHandle						listHandle,
		short							penMode
	)
{
#if 1
#pragma unused (listHandle, penMode)
#else
		Rect							viewRect;
		PenState						savePenState;
				
		viewRect = (**listHandle).rView;
		InsetRect(&viewRect, -4, -4);
		GetPenState(&savePenState);
		PenNormal();
		PenMode(penMode);
		PenSize(2, 2);
		FrameRect(&viewRect);
		SetPenState(&savePenState);
#endif
}

/*
 * DisposeList
 * Dispose of a list. Note: if the list content contains a handle,
 * some other function must dispose of the content first.
 */
void
DisposeList(
		ListHandle				*listHandlePtr
	)
{
		ListHandle				theList;
		
		if ((theList = *listHandlePtr) != NULL) {
			*listHandlePtr = NULL;
			LDispose(theList);
			FailMemError();
		}
}

/*
 * HideList
 * Move the list offscreen.
 */
void
HideList(
		ListHandle				listHandle
	)
{
		Rect					viewRect;
		
		viewRect = (**listHandle).rView;
		if (viewRect.left < 4096) {
			DrawFocusRect(listHandle, patBic);
			MoveList(listHandle, viewRect.left + 4096, viewRect.top + 4096);
		}
}

/*
 * ShowList
 * Move the list onscreen. (The focus rect is cleared).
 */
void
ShowList(
		ListHandle				listHandle
	)
{
		Rect					viewRect;
		
		viewRect = (**listHandle).rView;
		if (viewRect.left >= 4096) {
			MoveList(listHandle, viewRect.left - 4096, viewRect.top - 4096);
			DrawFocusRect(listHandle, patBic);
		}
}


/*
 * Reposition the list and its vertical scroll bar.
 */
void
MoveList(
		ListHandle						listHandle,
		short							leftEdge,
		short							topEdge
	)
{
		Rect							viewRect;
#define LIST	(**listHandle)
		
		viewRect = LIST.rView;
		if (viewRect.left != leftEdge || viewRect.top != topEdge) {
			InsetRect(&viewRect, -1, -1);
			InvalRect(&viewRect);
			OffsetRect(
				&LIST.rView,
				leftEdge - LIST.rView.left,
				topEdge - LIST.rView.top
			);
			viewRect = LIST.rView;
			InsetRect(&viewRect, -1, -1);
			InvalRect(&viewRect);
			MoveControl(
				LIST.vScroll,
				LIST.rView.right,
				LIST.rView.top - 1
			);
		}
#undef LIST
}
