/*								ManageSDPPanel.c								*/
/*
 * DTS Catalog Peek
 * ManageSDPPanel.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * These functions connect the real world to the Standard Directory Panel.
 *
 * Globals
 *	DBR.userIdentity
 *			The current user identity.
 * See UnpackSDPPanelSelection for an explanation of other panel-related globals.
 */
#include "DTSCatalogPeek.h"
#pragma segment MainCode

/*
 * CreateSDPDirectoryPanel
 * Create a Standard Directory Panel. Note that we create it initially invisible.
 * The window creation function (that called us) will call SDPShowPanel when it
 * has finished its work; this eliminates an extra panel redraw.
 */
SDPPanelHandle
CreateSDPDirectoryPanel(
		register CatalogBrowserPtr		dbp,
		const Rect						*viewRect,
		const PackedRLIPtr				initialRLI
	)
{
		OSErr					status;
		FontData				fontData;
		Str255					work;
		long					fontSize;
		SDPPanelHandle			result;
		
		/*
		 * Create the panel in Palatino 12 (or whatever)
		 */
		GetIndString(work, STRN_Parameters, kPanelFontSize);
		StringToNum(work, &fontSize);
		GetFontData(&fontData, kPanelFontName, fontSize);
		TextFont(fontData.fontNumber);
		TextSize(fontData.fontSize);
		status = SDPNewPanel(
				&result,				/* result				*/
				WINDOW,					/* In this window		*/
				viewRect,				/* Where in window		*/
				FALSE,					/* Initially invisible	*/
				TRUE,					/* Is enabled			*/
				initialRLI,				/* Initial RLI			*/
				NULL,					/* No typesList			*/
				0,						/* No typesList			*/
				DBR.userIdentity,		/* Current identity		*/
				kEnumAllMask,			/* All possible			*/
				kMatchAll,				/* All possible			*/
				(long) dbp				/* refCon -> DBR		*/
			);
		FailOSErr(status);
#ifdef kSDPInstallPanelBusyProc			/* TEMP	#ifdef			*/
		status = SDPInstallPanelBusyProc(
				result,
				AnimateVampDisplay
			);
		LOG(status, "\pSDPInstallPanelBusyProc");
#endif
		return (result);
}

/*
 * DisposeSDPDirectoryPanel
 * Dispose of the panel and any current selection.
 */
void
DisposeSDPDirectoryPanel(
		register CatalogBrowserPtr		dbp,
		SDPPanelHandle					*thePanel
	)
{
		SDPPanelHandle				panelHandle;
		
		if ((panelHandle = *thePanel) != NULL) {
			DisposeSDPSelection(dbp);
			*thePanel = NULL;
			FailOSErr(SDPDisposePanel(panelHandle));
		}
}


/*
 * DoPanelEvent
 * This function is called each time through the event loop.
 * SDPPanelEvent handles changes in the panel state (enable
 * events, and similar), and returns a "significant event" flag
 * if the user changes the selection.
 */
void
DoSDPPanelEvent(
		register CatalogBrowserPtr		dbp
	)
{
		OSErr					status;
		SDPPanelState			whatHappened;
		SDPPanelHandle			thePanel;

		if (DBR.panelHandle != NULL && (**DBR.panelHandle).visible)
			thePanel = DBR.panelHandle;
		else if (DBR.pdPanelHandle != NULL && (**DBR.pdPanelHandle).visible)
			thePanel = DBR.pdPanelHandle;
		else {
			thePanel = NULL;
		}
		if (thePanel != NULL) {
			status = SDPPanelEvent(
						thePanel,
						&EVENT,
						&whatHappened
					);
			FailOSErr(status);
			if (whatHappened == kSDPChangedSelection)
				DoChangedSDPSelection(dbp, thePanel);
		}
}

/*
 * DoChangedSDPSelection
 * The user has selected (or unselected) a directory entry.
 * Get the new selection.
 */
void
DoChangedSDPSelection(
		register CatalogBrowserPtr		dbp,
		SDPPanelHandle					thePanel
	)
{
		SDPSelectionState		selectionState;
		static const RecordID	nullRecordID;
		
		FailOSErr(
			SDPGetPanelSelectionState(
				thePanel,
				&selectionState
			)
		);
		/*
		 * If you do this, you can get invalid DSSpec and/or
		 * record id data - I don't understand this completely.
		 */
		if (selectionState != kSDPNothingSelected) {
			SetCursor(*GetCursor(watchCursor));
			/*
			 * We have a new selection.  If there is
			 * a current selection, dispose of it now.
			 * This closes the personal address book,
			 * if it was open.
			 */
			DisposeSDPSelection(dbp);
			UnpackSDPPanelSelection(dbp, thePanel);
			if (selectionState == kSDPRecordSelected
			 || selectionState == kSDPRecordAliasSelected)
				OpenPersonalAddressBook(dbp);
			InitCursor();
		}
		else {
			/*
			 * Nothing is selected, clear out the DSSpec,
			 * if there is one.
			 */
			DisposeSDPSelection(dbp);
		}
		ClearRecordDisplay(dbp);
		MakeCatalogDisplayList(dbp);
		if (DBR.selectedDSSpec != NULL
		 && DBR.isValidDSSpec
		 && DBR.specType == kOCERecordDSSpec) {
			EnumerateAttributeTypes(
				dbp,
				&DBR.panelRecordInfo
			);
		}
		/*
		 * Do this last so all the record stuff is
		 * correctly setup (DirGetDirectoryInfo needs
		 * the current record info).
		 */
		GetSDPIcon(dbp);
}

/*
 * DisposeSDPSelection
 * Dispose of all data relating to the current selection,
 * if any is present, of course.
 */
void
DisposeSDPSelection(
		register CatalogBrowserPtr		dbp
	)
{
		ForgetPtr(DBR.selectedDSSpec);
		ClosePersonalAddressBook(dbp);
		DBR.theDSSpec = gNullDSSpec;
		DBR.panelRecordInfo.recordID = gNullRecordID;
		DBR.theDSSpecRLI = gNullRLI;
		DBR.isValidDSSpec = FALSE;
		ClearRecordDisplay(dbp);
}

