/*									TwistDownList.c								*/
/*
 * List In A List Sample
 * TwistDownList.c
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * TwistDownList manages all aspects of the "twist-down" list, a one-column text
 * list where each row contains a triangular button. The button may be in one of
 * two states: closed and opened. When opened, sub-elements to this row are
 * displayed. The twist-down list is generally based on code from NewsWatcher
 * by Steve Falkenburg, but Steve claims that the code was originally written
 * by John Norstad.
 */
#include "TwistDownList.h"
#include <Script.h>
#include <Palettes.h>
#ifndef THINK_C
#include <SysEqu.h>
#endif
#define IS_COLOR(port) (((((CGrafPtr) (port))->portVersion) & 0xC000) != 0)

/*
 * kAnimationDelay is the number of ticks to display the intermediate
 * "twisting" glyph.
 */
#define kAnimationDelay		3L
/*
 * This defines the amount of space to display to the left
 * and right of the twist-down triangle.
 */
#define kTriangleLeftGap		(1)
#define kTriangleRightGap		(1)
#define kScrollBarWidth			(16)

/*
 * The List's userHandle contains some private context information.
 * The PolyHandles are used to draw the "triangle" buttons. Note that they were
 * drawn to the list cell height. If that should change, the triangles will
 * be reconstructed. fontSize, fontNumber, and isLeftJustify are used to
 * draw the list cell content.
 */
typedef struct TwistDownPrivateRecord {
		TwistDownDrawProc	drawProc;				/* This draws the cell data	*/
		PolyHandle			openTriangle;			/* The "expanded" button	*/
		PolyHandle			collapsedTriangle;		/* The "closed" button		*/
		PolyHandle			intermediateTriangle;	/* The "expanding" button	*/
		short				tabIndent;				/* NewTwistDownList param	*/
		short				cellHeight;				/* List cell height			*/
		short				fontSize;				/* for TextSize				*/
		short				fontNumber;				/* for TextFont				*/
		Boolean				canHiliteSelection;		/* TRUE if hilite ok		*/
		Boolean				isLeftJustify;			/* GetSysJust value			*/
		short				triangleWidth;			/* Twist-down button width	*/
} TwistDownPrivateRecord, *TwistDownPrivatePtr, **TwistDownPrivateHdl;

/*
 * All functions use these macros to access internal data:
 *	LIST				The de-referenced list handle.
 *	PRIVATE				The de-referenced private data record
 *	ELEM				A de-referenced list element.
 */
#define LIST			(**theList)
#define PRIVATE			(**((TwistDownPrivateHdl) (LIST.userHandle)))
#define ELEM			(**twistDownHdl)
#define FLAG			(ELEM.flag)


/*
 * This code sequence is copied into the ListProc handle. It is designed so we
 * don't have to flush the instruction and data caches. Note: this must be revised
 * for Power PC.
 */
static const short gDummyLDEF[] = {
		0x207A,			/*		movea.l	procPtr,a0			*/
		0x0004,			/*				<offset to procPtr>	*/
		0x4ED0,			/*		jmp		a0 					*/
		0x0000, 0x0000	/*		dc.l	<Drawing Proc here>	*/
};

/*
 * Local (private) functions.
 */
/*
 * Dispose of a PolyHandle. This must be a macro. The argument
 * may not have side-effects.
 */
#define ForgetPoly(thePoly) do {		\
		if (thePoly != NULL) {	 		\
			KillPoly(thePoly);			\
			thePoly = NULL;				\
		}								\
	} while (0)
#define height(r)		((r).bottom - (r).top)
#define width(r)		((r).right - (r).left)

static short					CountVisibleElements(
		TwistDownHdl				twistDownHdl
	);
static void						SetElementsInList(
		ListHandle					theList,
		TwistDownHdl				twistDownHdl,
		Cell						*currentCell
	);
static pascal void				TwistDownLDEF(
		short						listMessage,
		Boolean						listSelect,
		Rect						*listRect,
		Cell						listCell,
		short						listDataOffset,
		short						listDataLen,
		ListHandle					listHandle
	);
static void						DrawTriangle(
		WindowPtr					theWindow,
		PolyHandle					polyHandle,
		Point						polyPoint,
		Boolean						isSelected
	);
static void						CreateTriangles(
		ListHandle					theList
	);

/*
 * Create a twist-down list. Before calling, you must set the port to the current
 * window and specify the font and font size that is to be used to draw the list.
 * NewTwistDownList creates an empty one-column list with a vertical scroll bar
 * and no grow box. Only one item may be selected at a time; but this could
 * be changed by the application without difficulty.
 *
 * Note: unlike LNew, the viewRect includes the vertical scroll bar. The frame
 * and focus rectangles should be drawin outside of the list's rView.
 *
 * The tabIndent parameter should be set to the amount to indent successive levels
 * (zero means no indentation). Setting it to the widMax value from the current
 * font seems reasonable.
 *
 * canHiliteSelection should be TRUE for normal selection (the selection is
 * hilited). It should be FALSE if you want to supress selection. Because
 * hirearchical lists often "select" by revealing a sub-topic, this may be
 * more reasonable in many cases.
 */
ListHandle
NewTwistDownList(
		const Rect					*viewRect,
		TwistDownDrawProc			drawProc,
		unsigned short				tabIndent,
		Boolean						canHiliteSelection
	)
{
		register TwistDownPrivatePtr	privatePtr;
		ListHandle						theList;
		short							listHeight;
		Point							cellSize;
		Rect							dataBounds;
		Rect							listRect;
		Handle							drawProcHdl;
		FontInfo						info;
		short							listFontHeight;
		OSErr							status;
		ProcPtr							listDefProc;

		GetFontInfo(&info);
		listFontHeight = info.ascent + info.descent + info.leading;
		/*
		 * Define the list drawing area, adjusting the list area height so an
		 * integral number of rows will be drawn. 
		 */
		listRect = *viewRect;
		listRect.right -= kScrollBarWidth;
		listHeight = height(listRect);
		listHeight -= (listHeight % listFontHeight);
		listRect.bottom = listRect.top + listHeight;
		SetPt(&cellSize, width(listRect), listFontHeight);
		/*
		 * Define a one-column list.
		 */
		SetRect(&dataBounds, 0, 0, 1, 0);
		theList = LNew(
				&listRect,				/* Viewing area							*/
				&dataBounds,			/* Rows and col's						*/
				cellSize,				/* Element size							*/
				0,						/* No defProc yet						*/
				qd.thePort,				/* Display window						*/
				TRUE,					/* Draw it								*/
				FALSE,					/* No grow box							*/
				FALSE,					/* No horiz scroll						*/
				TRUE					/* Vertical scroll						*/
			);
		if (theList == NULL)
			goto exit;
		LIST.selFlags = lOnlyOne;		/* Allow a single selection				*/
		LIST.userHandle = NewHandleClear(sizeof (TwistDownPrivateRecord));
		LIST.refCon = 0;				/* Paranoia								*/
		if (LIST.userHandle == NULL)
			goto failure;
		privatePtr = (TwistDownPrivatePtr) (*LIST.userHandle);
#define PRIV (*privatePtr)
		PRIV.drawProc = drawProc;
		PRIV.tabIndent = tabIndent;
		PRIV.canHiliteSelection = canHiliteSelection;
		PRIV.fontNumber = qd.thePort->txFont;
		PRIV.fontSize = qd.thePort->txSize;
#undef PRIV
		/*
		 * Note that we left PRIV.cellHeight set to zero. The first time the LDEF
		 * is called, it will detect that the cell size has changed and will
		 * create properly-sized twist-down button PolyHandles.
		 */
		/*
		 * Create a handle with a callback to our private drawing function.
		 */
		status = PtrToHand(gDummyLDEF, &drawProcHdl, sizeof gDummyLDEF);
		if (status != noErr)
			goto failure;
		listDefProc = (ProcPtr) TwistDownLDEF;
		BlockMove(
			(Ptr) &listDefProc,
			&((short *) *drawProcHdl)[3],
			sizeof listDefProc
		);
		LIST.listDefProc = drawProcHdl;
		goto exit;
failure:
		if (theList != NULL)
			LDispose(theList);
		theList = NULL;
exit:	return (theList);
}


/*
 * DoTwistDownClick handles all processing after a click in the list window.
 * It returns an indication of the user action:
 *	kTwistDownNotInList
 *		The click was not in this list. Your application may ignore this click or
 *		take other appropriate action.
 *	kTwistDownNoClick
 *		The user released the mouse outside of the list area: this click should be
 *		ignored. (It may have been a click in the scroll bar.)
 *	kTwistDownButtonClick
 *		The user clicked on the twist-down button. The list will be expanded or
 *		contracted as appropriate.
 *	kTwistDownClick
 *		The user clicked (once) on a list datum. The application should treat this
 *		as an item selection.
 *	kTwistDownDoubleClick
 *		The user double-clicked on a list datum. The application should open this
 *		item or take other appropriate action.
 *
 * If DoTwistDownClick returns kTwistDownButtonClick, kTwistDownClick, or
 * kTwistDownDoubleClick, selectedListCell will be set to the cell that the user
 * clicked on.
 */
TwistDownClickState
DoTwistDownClick(
		ListHandle					theList,
		EventRecord					*eventRecordPtr,
		Cell						*selectedListCell
	)
{
		Cell						theCell;
		Rect						hitRect;
		Boolean						inHitRect;
		Boolean						newInHitRect;
		short						cellHeight;
		short						visibleTop;
		short						cellSize;
		TwistDownHdl				twistDownHdl;
		Point						penPt;
		TwistDownClickState			result;
		long						finalTicks;

		penPt = eventRecordPtr->where;
		GlobalToLocal(&penPt);
		hitRect = LIST.rView;
		hitRect.right += kScrollBarWidth;
		if (PtInRect(penPt, &hitRect) == FALSE)
			result = kTwistDownNotInList;
		else {
			/*
			 * Set hitRect to the area of the list that contains the twist-down
			 * triangle button. Note that this presumes a list where only column
			 * zero is displayed.
			 */
			CreateTriangles(theList);
			if (PRIVATE.isLeftJustify) {
				hitRect.right = LIST.rView.left
						+ LIST.indent.h
						+ PRIVATE.triangleWidth;
			}
			else {
				hitRect.left = LIST.rView.right
						- LIST.indent.h
						- PRIVATE.triangleWidth;
			}
			inHitRect = FALSE;
			if (PtInRect(penPt, &hitRect)) {
				/*
				 * It's in a the triangle area. Find the selected cell and check
				 * whether this cell's element has a visible twist-down button.
				 */
				visibleTop = LIST.visible.top;
				cellHeight = LIST.cellSize.v;
				theCell.h = 0;		/* This has the visual content			*/
				theCell.v =
					((penPt.v - LIST.rView.top) / cellHeight)
					+ visibleTop;
				if (theCell.v < LIST.dataBounds.bottom) {
					/*
					 * There should be an element index in this cell.
					 * Grab it and make sure that this element actually
					 * has a sub-list. If not, just ignore the click.
					 */
					cellSize = sizeof twistDownHdl;
					LGetCell(&twistDownHdl, &cellSize, theCell, theList);
					/*
					 * set inHitRect TRUE if there is a sub-list button here.
					 * Note: it is possible to have a button but no actual sub-list
					 * (consider an empty folder in a disk hierarchy: the presence
					 * of the button indicates "this is a folder" to the user).
					 */
					if (cellSize == sizeof twistDownHdl && FLAG.hasTwistDown)
						inHitRect = TRUE;
				}
			}
			if (inHitRect == FALSE) {
				/*
				 * This cell doesn't have an expansion triangle, or the user did
				 * not click in the button area. Just call the  normal list click
				 * handler to manage the scroll bars. Set result appropriately.
				 */
				result = (LClick(penPt, eventRecordPtr->modifiers, theList))
					? kTwistDownDoubleClick
					: kTwistDownClick;
			}
			else {
				/*
				 * The user clicked on in the twist-down button area. Simulate a
				 * button click and track the mouse as it wanders in and out of the
				 * button area. (inHitRect is true at this point). Whenever the
				 * button selection state changes, call LDraw to redraw the
				 * twist-down triangle. This is the way to simulate TrackControl.
				 */
				FLAG.drawButtonFilled = TRUE;
				FLAG.onlyRedrawButton = TRUE;
				LDraw(theCell, theList);
				/*
				 * Set hitRect to the dimensions of the twist-down button.
				 */
				hitRect.top =
					((theCell.v - visibleTop) * cellHeight)
					+ LIST.rView.top;
				hitRect.bottom = hitRect.top + cellHeight;
				/*
				 * Track the mouse while it still down: if it moves into the
				 * triangle rectangle, redraw it filled, if it moves out of the
				 * triangle, redraw it unfilled.
				 */
				if (StillDown()) {
					while (WaitMouseUp()) {
						GetMouse(&penPt);
						newInHitRect = PtInRect(penPt, &hitRect);
						if (newInHitRect != inHitRect) {
							/*
							 * The mouse moved in or out of the triangle.
							 */
							FLAG.drawButtonFilled = newInHitRect;
							LDraw(theCell, theList);
							inHitRect = newInHitRect;
						}
					}
				}
				/*
				 * The user released the mouse.
				 */
				if (inHitRect == FALSE) {
					/*
					 * Normally, drawButtonFilled will be clear. It can be set,
					 * however, if the user clicks so briefly on the triangle that
					 * the StillDown() test above is FALSE.
					 */
					if (FLAG.drawButtonFilled) {
						FLAG.drawButtonFilled = FALSE;
						LDraw(theCell, theList);
					}
					result = kTwistDownNoClick;
				}
				else {
					/*
					 * The user released the mouse in the expansion triangle.
					 * Draw an intermediate "animation" triangle. Then call 
					 * ExpandOrCollapseTwistDownList which will redraw the
					 * button in its new state.
					 */
					FLAG.drawIntermediate = TRUE;
					LDraw(theCell, theList);
					Delay(kAnimationDelay, &finalTicks);
					FLAG.drawIntermediate = FALSE;
					FLAG.drawButtonFilled = FALSE;
					ExpandOrCollapseTwistDownList(theList, theCell);
					result = kTwistDownButtonClick;
					*selectedListCell = theCell;
				}
				FLAG.onlyRedrawButton = FALSE;
			}
		}
		return (result);
}

/*
 * ExpandOrCollapseSubList modifies the "show sublist" flag for the selected
 * cell and rebuilds the visual display.
 */
void
ExpandOrCollapseTwistDownList(
		ListHandle					theList,
		Cell						selectedListCell
	)
{
		TwistDownHdl				twistDownHdl;
		
		twistDownHdl = GetTwistDownElementHandle(theList, selectedListCell);
		if (twistDownHdl != NULL) {
			if (FLAG.hasTwistDown) {
				FLAG.showSublist = !(FLAG.showSublist);
				/*
				 * Redraw the twist-down button in its new state.
				 */
				FLAG.drawButtonFilled = FALSE;
				FLAG.onlyRedrawButton = TRUE;
				LDraw(selectedListCell, theList);
				FLAG.onlyRedrawButton = FALSE;
				/*
				 * If some other part of the list will change, rebuild the
				 * List Manager list cells and redraw the list.
				 */
				if (ELEM.subElement != NULL)
					BuildVisibleList(theList, selectedListCell.v);
			}
		}
}

/*
 * CreateVisibleList is called when the list is created. It stores the list
 * head into cell [0, 0] and calls BuildVisibleList to instantiate the display.
 */
void
CreateVisibleList(
		ListHandle					theList,
		TwistDownHdl				twistDownHdl
	)
{
		Cell						theCell;
		
		if (LIST.dataBounds.bottom == 0) {
			/*
			 * Add one row to the list so there is a place to put the head element.
			 */
			LDoDraw(FALSE, theList);
			LAddRow(1, 0, theList);
			LDoDraw(TRUE, theList);
		}
		SetPt(&theCell, 0, 0);
		LSetCell(&twistDownHdl, sizeof twistDownHdl, theCell, theList);
		BuildVisibleList(theList, 0);
}

/*
 * BuildVisibleList is called when the list is created, or when the user clicks on
 * a twist-down button. selectedRow is the first row that needs to be redrawn.
 * To rebuild the entire list, store the list head into cell [0, 0] and call
 * with selectedRow = 0. Note: because people can be rebuilding the list from
 * within a larger sublist context, the list head must be stored in cell [0, 0].
 */
void
BuildVisibleList(
		ListHandle					theList,
		short						selectedRow
	)
{
		short						nRows;			/* How many we need	to show	*/
		short						currentRows;	/* How many are in the list	*/
		Rect						viewRect;
		TwistDownHdl				twistDownHdl;
		TwistDownHdl				listHead;
		Cell						theCell;
		
		LDoDraw(FALSE, theList);
		SetPt(&theCell, selectedRow, 0);
		twistDownHdl = GetTwistDownElementHandle(theList, theCell);
		SetPt(&theCell, 0, 0);
		listHead = GetTwistDownElementHandle(theList, theCell);
		nRows = CountVisibleElements(listHead);
		currentRows = LIST.dataBounds.bottom;
		if (currentRows > nRows)
			LDelRow(currentRows - nRows, nRows, theList);	/* Shrink the list	*/
		else if (currentRows < nRows)
			LAddRow(nRows - currentRows, currentRows + 1, theList);	/* Grow it	*/
		if (nRows != 0) {
			SetElementsInList(theList, listHead, &theCell);
		}
		LDoDraw(TRUE, theList);
		/*
		 * Redraw any elements that are greater than the inserted row.
		 */
		viewRect = LIST.rView;
		viewRect.top += ((selectedRow + 1) - LIST.visible.top) * LIST.cellSize.v;
		if (viewRect.top < viewRect.bottom)
			InvalRect(&viewRect);
}

/*
 * CountVisibleElements is a recursive function that returns the number of visible
 * elements in the list. Call with the list head to process the entire list. Note:
 * list elements are visible, but sub-lists are visible only if the button state
 * requests visiblity. This function may be called with a NULL argument without
 * problems. Also note that we ignore the current List Manager cell data, working
 * only with the actual linked list structure. By convention, however, the list
 * head is stored in cell [0, 0].
 */
static short
CountVisibleElements(
		TwistDownHdl				twistDownHdl
	)
{
		short						result;

		result = 0;
		while (twistDownHdl != NULL) {
			++result;
			if (FLAG.showSublist)
				result += CountVisibleElements(ELEM.subElement);
			twistDownHdl = ELEM.nextElement;
		}			
		return (result);
}

/*
 * SetElementsInList is a recursive function that copies visible list elements from
 * the linked list to the List Manager list. The List Manager list was extended so
 * it holds all visible cells. Note that currentCell is a "global" that always
 * contains the current List Manager cell. To process the entire list, set
 * currentCell to [0, 0] and call with the list head.
 */ 
static void
SetElementsInList(
		ListHandle					theList,
		TwistDownHdl				twistDownHdl,
		Cell						*currentCell
	)
{
		while (twistDownHdl != NULL) {
			LSetCell(&twistDownHdl, sizeof twistDownHdl, *currentCell, theList);
			++currentCell->v;
			if (FLAG.showSublist)
				SetElementsInList(theList, ELEM.subElement, currentCell);
			twistDownHdl = ELEM.nextElement;
		}
}

/*
 * GetTwistDownElementHandle returns the TwistDownHdl that is stored in a List
 * cell. It will return NULL if the cell is out of bounds.
 */
TwistDownHdl
GetTwistDownElementHandle(
		ListHandle					theList,
		Cell						theCell
	)
{
		TwistDownHdl				twistDownHdl;
		short						dataSize;
		
		dataSize = sizeof twistDownHdl;
		LGetCell(&twistDownHdl, &dataSize, theCell, theList);
		if (dataSize != sizeof twistDownHdl)
			twistDownHdl = NULL;
		return (twistDownHdl);
}

/*
 * MakeTwistDownElement adds an element to a linked list. It is designed to create
 * elements in a hierarchical linked-list where each element may be followed by a
 * successor (on the same "level") and/or by a child list (on a lower "level").
 * Note that MakeTwistDownElement is only concerned with the linked list. It
 * does not change the List Manager list.
 *
 * The parameters are as follows:
 *	previousIndex
 *		This has a handle to of the predecessor to this element.
 *		If this is the first element in a list, or the first on a level,
 *		previousIndex should be set to NULL.
 *	indentLevel
 *		This is the indentation-level of this list element. It is only
 *		used to tab the list elements on the visual display. If you don't
 *		want tabbing, set tabIndent to zero when the list was created.
 *	dataLength
 *		This is the length of the list element datum.
 *	dataPtr
 *		This is a pointer to the first byte of the list element datum. If NULL,
 *		a data block of the requisite size will be created, but the caller is
 *		responsible for filling it in.
 *	result
 *		If MakeElement succeeds, result will will contain a handle to the
 *		list element it created. This is needed to create a successor
 *		element.
 */
OSErr
MakeTwistDownElement(
		TwistDownHdl				previousElement,
		short						indentLevel,
		unsigned short				dataLength,
		Ptr							dataPtr,
		TwistDownHdl				*result
	)
{
		TwistDownHdl				twistDownHdl;
		OSErr						status;
		static const TwistDownRecord nullElement;
		
		twistDownHdl = (TwistDownHdl) NewHandle(
							sizeof (TwistDownRecord)
							+ dataLength
							- 1
						);
		status = MemError();
		if (status == noErr) {
			if (previousElement != NULL)
				(**previousElement).nextElement = twistDownHdl;
			ELEM = nullElement;
			ELEM.indentLevel = indentLevel;
			ELEM.dataLength	= dataLength;
			if (dataPtr != NULL)
				BlockMove(dataPtr, &ELEM.data[0], dataLength);
			*result = twistDownHdl;
		}
		return (status);
}

/*
 * DisposeTwistDownHdl disposes of the linked list that has the argument at
 * its head, and of all sublists linked to this list. Note that it presumes
 * that the list element does not, itself, contain Ptr or Handle data that
 * must be disposed. (C++ overloading would be useful here.)
 */
void
DisposeTwistDownHdl(
		TwistDownHdl				twistDownHdl
	)
{
		TwistDownHdl				nextElement;
		TwistDownHdl				subElement;
		
		while (twistDownHdl != NULL) {
			nextElement = ELEM.nextElement;
			subElement = ELEM.subElement;
			DisposeHandle((Handle) twistDownHdl);
			if (subElement != NULL)
				DisposeTwistDownHdl(subElement);
			twistDownHdl = nextElement;
		}
}

/*
 * Draw the twist-down list cell. Note that we have to draw the expansion button
 * in one of five states: (expanded/compressed), (normal/selected), and an
 * animation state.
 */
pascal void
TwistDownLDEF(
		short							listMessage,
		Boolean							listSelect,
		Rect							*listRect,
		Cell							listCell,
		short							listDataOffset,
		short							listDataLen,
		ListHandle						theList
	)
{
#pragma unused (listCell, listDataOffset)

		short							indent;
		TwistDownHdl					twistDownHdl;
		register TwistDownPtr			twistDownPtr;
		short							cellSize;
		PolyHandle						polyHandle;
		Point							polyPoint;
		Rect							viewRect;
		signed char						elementLockState;

#define LELEM	(*twistDownPtr)
#define LFLAG	(LELEM.flag)

		switch (listMessage) {
		case lInitMsg:
			break;
		case lCloseMsg:
			if (LIST.userHandle != NULL) {
				ForgetPoly(PRIVATE.openTriangle);
				ForgetPoly(PRIVATE.collapsedTriangle);
				ForgetPoly(PRIVATE.intermediateTriangle);
				DisposeHandle(LIST.userHandle);
				LIST.userHandle = NULL;
			}
			break;
		case lDrawMsg:
			if (listDataLen > 0 && LIST.userHandle != NULL) {
				CreateTriangles(theList);
				/*
				 * Get the cell content. This is the handle that has the list
				 * element. We check that the userHandle has been setup correctly.
				 * Note that we don't use LFind (or similar) because the data might
				 * not be aligned in the list cell storage. Actually, the data is
				 * aligned as we only store Handles in the cells.
				 */
				cellSize = sizeof twistDownHdl;
				LGetCell(&twistDownHdl,	&cellSize, listCell, theList);
				if (cellSize == sizeof twistDownHdl && twistDownHdl != NULL) {
					elementLockState = HGetState((Handle) twistDownHdl);
					HLock((Handle) twistDownHdl);
					twistDownPtr = (*twistDownHdl);
					viewRect = *listRect;
					if (LFLAG.onlyRedrawButton) {
						if (PRIVATE.isLeftJustify) {
							viewRect.right = viewRect.left
										+ LIST.indent.h
										+ PRIVATE.triangleWidth;
						}
						else {
							viewRect.left = viewRect.right
									- LIST.indent.h
									- kTriangleLeftGap
									- PRIVATE.triangleWidth;
						}
					}
					EraseRect(&viewRect);
					if (LFLAG.hasTwistDown) {
						/*
						 * Draw the expansion triangle in one of
						 * its three states.
						 */
						polyPoint.v = listRect->top + 1;
						if (PRIVATE.isLeftJustify) {
							polyPoint.h = listRect->left
										+ LIST.indent.h
										+ kTriangleLeftGap;
						}
						else {
							polyPoint.h = listRect->right
										- LIST.indent.h
										- PRIVATE.triangleWidth
										+ kTriangleRightGap;
						}
						if (LFLAG.drawIntermediate)
							polyHandle = PRIVATE.intermediateTriangle;
						else if (LFLAG.showSublist)
							polyHandle = PRIVATE.openTriangle;
						else {
							polyHandle = PRIVATE.collapsedTriangle;
						}
						DrawTriangle(
							LIST.port,
							polyHandle,
							polyPoint,
							LFLAG.drawButtonFilled
						);
					}
					if (LFLAG.onlyRedrawButton == FALSE
					 && LELEM.dataLength > 0) {
						/*
						 * Indent the text to show the depth of the hierarchy.
						 */
						indent = LIST.indent.h
							+ PRIVATE.triangleWidth
							+ (PRIVATE.tabIndent * LELEM.indentLevel);
						viewRect = *listRect;
						/*
						 * Build a display rectangle for the cell text and
						 * set the pen to the leftmost position of the text.
						 * Note that this right-justifies text for Arabic
						 * and Hebrew.
						 */
						if (PRIVATE.isLeftJustify)
							viewRect.left += indent;
						else {
							viewRect.right -= indent;
						}
						if (PRIVATE.drawProc == NULL) {
							if (PRIVATE.isLeftJustify)
								MoveTo(viewRect.left, viewRect.top + LIST.indent.v);
							else {
								MoveTo(
									viewRect.right
									- TextWidth(LELEM.data, 0, LELEM.dataLength),
									viewRect.top + LIST.indent.v
								);
							}
							DrawText(LELEM.data, 0, LELEM.dataLength);
						}
						else {
							(PRIVATE.drawProc)(
								theList,
								(const Ptr) LELEM.data,
								LELEM.dataLength,
								&viewRect
							);
						}
					}							/* Drawing cell					*/
					HSetState((Handle) twistDownHdl, elementLockState);
				}								/* Have list element			*/
			}									/* Have cell data				*/
			if (listSelect == FALSE)
				break;
			/* Continue to do hilite */
		case lHiliteMsg:
			if (PRIVATE.canHiliteSelection) {
#ifdef THINK_C
				HiliteMode &= ~(1 << hiliteBit);
#else /* MPW */
				*((char *) HiliteMode) &= ~(1 << hiliteBit);	/* IM V-61	*/
#endif
				viewRect = *listRect;
				if (PRIVATE.isLeftJustify)
					viewRect.left += (LIST.indent.h + PRIVATE.triangleWidth);
				else {
					viewRect.right -= (LIST.indent.h + PRIVATE.triangleWidth);
				}
				InvertRect(&viewRect);
			}
			break;
		}
#undef LELEM
#undef LFLAG
}

/*
 * DrawTriangle
 * Draw the polygon and fill it so it looks a bit like the Finder. If isSelected
 * is TRUE, the polygon is always black-filled. Else, on color monitors, a light
 * gray is chosen.
 */
static void
DrawTriangle(
		WindowPtr					theWindow,
		PolyHandle					polyHandle,
		Point						polyPoint,
		Boolean						isSelected
	)
{
		RGBColor					foreColor;
		RGBColor					saveForeColor;
		RGBColor					backColor;
		short						i;
		
		OffsetPoly(polyHandle, polyPoint.h, polyPoint.v);
		if (isSelected)
			FillPoly(polyHandle, (ConstPatternParam) &qd.black);
		else {
			/*
			 * If this is a color window, fill the unselected triangle with a very
			 * light gray. The finder extends this by using the icon color instead
			 * of black. 
			 */
			if (IS_COLOR(theWindow)) {
				GetForeColor(&foreColor);
				saveForeColor = foreColor;
				GetBackColor(&backColor);
				for (i = 0; i < 8; i++) {
					/*
					 * Get the lightest gray that GetGray can differentiate.
					 */
					if (GetGray(GetGDevice(), &backColor, &foreColor) == FALSE)
						break;
				}
				RGBForeColor(&foreColor);
				FillPoly(polyHandle, (ConstPatternParam) &qd.black);
				RGBForeColor(&saveForeColor);
			}
		}
		FramePoly(polyHandle);
		OffsetPoly(polyHandle, -polyPoint.h, -polyPoint.v);
}

/*
 * CreateTriangles
 * This function creates the three states of the twist-down button that are drawn
 * in the display list.
 *
 * CreateTriangles is called when the list is drawn. If the list was newly created
 * or resized (or the window justification changed), the triangle buttons will
 * be created or re-created.
 *
 * Note that the port and text drawing characteristics must have been set.
 */
static void
CreateTriangles(
		ListHandle					theList
	)
{
		short						size;
		short						halfSize;
		short						intermediateSize;
		FontInfo					info;
		Boolean						currentLeftJustify;

		currentLeftJustify = (GetSysJust() == 0);
		if (LIST.cellSize.v != PRIVATE.cellHeight
		 || currentLeftJustify != PRIVATE.isLeftJustify) {
			PRIVATE.cellHeight = LIST.cellSize.v;
			PRIVATE.isLeftJustify = currentLeftJustify;
			ForgetPoly(PRIVATE.openTriangle);
			ForgetPoly(PRIVATE.collapsedTriangle);
			ForgetPoly(PRIVATE.intermediateTriangle);
			GetFontInfo(&info);
			size = info.ascent;					/* The "show sublist" button	*/
			size &= ~1;							/* Round down to an even number	*/
			halfSize = size / 2;
			intermediateSize = (size * 3) / 4;
			PRIVATE.openTriangle = OpenPoly();
				MoveTo(0, halfSize);
				LineTo(size, halfSize);
				LineTo(halfSize, size);
				LineTo(0, halfSize);
			ClosePoly();
			if (PRIVATE.isLeftJustify) {		/* Roman alphabet triangles		*/
				PRIVATE.collapsedTriangle = OpenPoly();
					MoveTo(halfSize, 0);
					LineTo(size, halfSize);
					LineTo(halfSize, size);
					LineTo(halfSize, 0);
				ClosePoly();
				PRIVATE.intermediateTriangle = OpenPoly();	
					MoveTo(intermediateSize, 0);
					LineTo(intermediateSize, intermediateSize);
					LineTo(0, intermediateSize);
					LineTo(intermediateSize, 0);
				ClosePoly();
			}
			else {
				/*
				 * Triangles face the other way for right to left languages.
				 */
				PRIVATE.collapsedTriangle = OpenPoly();
					MoveTo(size - halfSize, 0);
					LineTo(0, halfSize);
					LineTo(size - halfSize, size);
					LineTo(size - halfSize, 0);
				ClosePoly();
				PRIVATE.intermediateTriangle = OpenPoly();
					MoveTo(size - intermediateSize, 0);
					LineTo(size - intermediateSize, intermediateSize);
					LineTo(size, intermediateSize);
					LineTo(size - intermediateSize, 0);
				ClosePoly();
			}		
			/*
			 * Remember the width of the "button" area.
			 */
			PRIVATE.triangleWidth =
						(**PRIVATE.openTriangle).polyBBox.right
						+ kTriangleLeftGap
						+ kTriangleRightGap;
		}
}
