/*									TwistDownList.h								*/
/*
 * List In A List Sample
 * TwistDownList.h
 * Copyright  1993 Apple Computer Inc. All rights reserved.
 *
 * Note that "list" is ambiguous, referring either to a List Manager list or to
 * a "Lisp-like" list. In the following documentation, "cell" will always refer
 * to a List Manager list cell, while "element" will always refer to a Lisp
 * cell. Note that elements are not always displayed by the List Manager.
 */
#ifndef _TwistDownList_
#define _TwistDownList_

#ifndef REZ
#ifndef THINK_C				/* MPW includes			*/
#include <Types.h>
#include <Lists.h>
#include <QuickDraw.h>
#include <Packages.h>
#include <Traps.h>
#include <Windows.h>
#endif
#ifndef TRUE
#define TRUE				1
#define FALSE				0
#endif
#ifndef NULL
#define NULL				0
#endif
/*
 * TwistDownClickState is the value returned by DoTwistDownClick.
 */
typedef enum {
	kTwistDownNotInList = 0,			/* The click was not in this list		*/
	kTwistDownNoClick,					/* User changed mind					*/
	kTwistDownButtonClick,				/* Clicked on twist-down button			*/
	kTwistDownClick,					/* Single-click on list datum			*/
	kTwistDownDoubleClick				/* Double-click on list datum			*/
} TwistDownClickState;

/*
 * This is the information needed to manage the twist-down display list. It is,
 * essentially, a classical Lisp list with recursively-defined sublists. Each
 * element contains enough information to display one component. This list is
 * stored in a handle stored in the list's private data handle. The ListManager
 * list cells contain TwistDownHdl pointers to the visible elements.
 * A value of NULL in nextElement indicates the end of a list or sublist.
 * A value of NULL in subElement means that there are no sub-elements (this
 * is a leaf). Note that the Macintosh memory manager handles garbage collection.
 *
 * The flag bits have the following meanings:
 *	drawTwistDown		This list element is a sub-list (it may be empty).
 *						Draw the twist-down button (even if subElement is NULL).
 *	showSublist			This list element is has a sub-list and the user wants
 *						to see the contents of the sub-list, too.
 * The following flags are only needed by the mouse-click handler.
 *	drawButtonFilled	This is used by the mouse-click handler to signal
 *						"mouse-down in button." The LDEF fills the button area.
 *	onlyRedrawButton	This is used by the mouse-click handler to signal the
 *						LDEF to draw only the button, but not the list content.
 *	drawIntermediate	The user is expanding/collapsing the folder. Draw
 *						the animation triangle.
 */
typedef struct TwistDownRecord {
	struct TwistDownRecord	**nextElement;		/* CDR - next on this "level"	*/
	struct TwistDownRecord	**subElement;		/* CAR - first in a sublist		*/
	short				indentLevel;			/* Indentation depth			*/
	struct {
		unsigned		hasTwistDown : 1;		/* Draw the triangle if set		*/
		unsigned		showSublist : 1;		/* Draw sub-hierarchy, too		*/
		unsigned		drawButtonFilled;		/* Mouse-down in button			*/
		unsigned		onlyRedrawButton : 1;	/* LDEF only redraws button		*/
		unsigned		drawIntermediate : 1;	/* Draw opening animation		*/
	} flag;
	unsigned short		dataLength;				/* Length of actual datum		*/
	unsigned char		data[1];				/* Datum to display				*/
} TwistDownRecord, *TwistDownPtr, **TwistDownHdl;

/*
 * This is the function that is called to draw the user-specified list cell
 * contents. The port and font have been properly initialized. If this parameter
 * is specified as NULL, DrawText will be used. Note that the application can use
 * the list's refCon to pass other data to the drawing procedure. Applications
 * should not have any other need to access the list handle or its contents.
 */
typedef pascal void		(*TwistDownDrawProc)(
		ListHandle			twistDownListHdl,	/* The list itself				*/
		const Ptr			dataPtr,			/* -> user data in record		*/
		unsigned short		dataLength,			/* length of user datum			*/
		const Rect			*viewRect			/* Draw in this area			*/
	);

/*
 * Create a twist-down list. Before calling, you must set the port to the current
 * window and specify the font and font size that is to be used to draw the list.
 * NewTwistDownList creates an empty one-column list with a vertical scroll bar
 * and no grow box. Only one item may be selected at a time; but this could
 * be changed by the application without difficulty.
 *
 * Note: unlike LNew, you should specify a correct viewRect: the scroll bars and
 * frame will be created inside the view rect.
 *
 * The TwistDownDrawProc will be called to draw a selection. If specified as
 * NULL, DrawText will be called.
 *
 * The tabIndent parameter should be set to the amount to indent successive levels
 * (zero means no indentation). Setting it to the widMax value from the current
 * font seems reasonable.
 *
 * canHiliteSelection should be TRUE for normal selection (the selection is
 * hilited). It should be FALSE if you want to supress selection. Because
 * hirearchical lists often "select" by revealing a sub-topic, this may be
 * more reasonable in many cases.
 */
ListHandle						NewTwistDownList(
		const Rect					*viewRect,
		TwistDownDrawProc			drawProc,
		unsigned short				tabIndent,
		Boolean						canHiliteSelection
	);
/*
 * DoTwistDownClick handles all processing after a click in the list window.
 * It returns an indication of the user action:
 *	kTwistDownNotInList
 *		The click was not in this list. Your application may ignore this click or
 *		take other appropriate action.
 *	kTwistDownNoClick
 *		The user released the mouse outside of the list area: this click should be
 *		ignored.
 *	kTwistDownButtonClick
 *		The user clicked on the twist-down button. Expand or contract the list as
 *		appropriate.
 *	kTwistDownClick
 *		The user clicked (once) on a list datum. The application should treat this
 *		as an item selection (or, it may be a click in the scroll bar). This is
 *		the result of a FALSE return from LClick.
 *	kTwistDownDoubleClick
 *		The user double-clicked on a list datum. The application should open this
 *		item or take other appropriate action. This is the result of a TRUE
 *		return from LClick.
 *
 * If DoTwistDownClick returns kTwistDownButtonClick, kTwistDownClick, or
 * kTwistDownDoubleClick, selectedListCell will be set to the cell that the user
 * clicked on.
 */
TwistDownClickState				DoTwistDownClick(
		ListHandle					theList,
		EventRecord					*eventRecordPtr,
		Cell						*selectedListCell
	);
/*
 * ExpandOrCollapseTwistdownSubList is called when the user clicks on the
 * twist-down button. index and selectedListCell are the values returned by
 * DoTwistDownClick.
 */
void							ExpandOrCollapseTwistDownList(
		ListHandle					theList,
		Cell						selectedListCell
	);
/*
 * CreateVisibleList is called when the list is created. It stores the list
 * head into cell [0, 0] and calls BuildVisibleList to instantiate the display.
 */
void							CreateVisibleList(
		ListHandle					theList,
		TwistDownHdl				twistDownHdl
	);
/*
 * BuildVisibleList is called when the list is created, or when the user clicks on
 * a twist-down button. selectedRow is the first row that needs to be redrawn. The
 * head of the entire list has been stored into cell [0, 0].
 */
void							BuildVisibleList(
		ListHandle					theList,
		short						selectedRow
	);
/*
 * GetTwistDownElementHandle returns the TwistDownHdl that is stored in a List
 * cell. It will return NULL if the cell is out of bounds.
 */
TwistDownHdl					GetTwistDownElementHandle(
		ListHandle					theList,
		Cell						theCell
	);

/*
 * MakeTwistDownElement adds an element to a linked list. It is designed to create
 * elements in a hierarchical linked-list where each element may be followed by a
 * successor (on the same "level") and/or by a child list (on a lower "level").
 * The parameters are as follows:
 *	previousElement
 *		This has a handle to the predecessor to this element. PreviousElement
 *		should be NULL if this is the first element.
 *	indentLevel
 *		This is the indentation-level of this list element. It is only used to tab
 *		the list elements on the visual display. If you never want tabbing, set
 *		tabIndent to zero when the list was created. Indents start at zero.
 *	dataLength
 *		This is the length of the list element datum.
 *	dataPtr
 *		This is a pointer to the first byte of the list element datum. If NULL,
 *		a data block of the requisite size will be created, but the caller is
 *		responsible for filling it in.
 *	result
 *		If MakeElement succeeds, result will will contain a handle to the
 *		list element it created. This is needed to create a successor
 *		element.
 */
OSErr							MakeTwistDownElement(
		TwistDownHdl				previousElement,
		short						indentLevel,
		unsigned short				dataLength,
		Ptr							dataPtr,
		TwistDownHdl				*result
	);

/*
 * This recursive function disposes of the argument list that has the argument
 * at its head, and of all sublists linked to this list.
 */
void
DisposeTwistDownHdl(
		TwistDownHdl				twistDownHdl
	);


#endif	/* REZ				*/
#endif	/* _TwistDownList_	*/
