/*
 * SimpleTextWindow.c
 * Copyright  1993, Apple Computer Inc. All Rights Reserved.
 * Output file writer for the Audit library.
 */
/*
 * Based on software
 * Copyright  1989-1992 Martin Minow.
 * All rights reserved.
 * Used by permission.
 */
/*
 * Usage:
 *		WindowPtr			NewTextWindow(
 *			StringPtr			windowTitle
 *			Rect				*boundsRect,
 *			Boolean				hasGoAway,
 *			short				fontNumber,
 *			short				fontSize
 *		);
 *		TextWindowState		TextWindowEvent(
 *			WindowPtr			textWindowPtr,
 *			EventRecord			*theEvent
 *		);
 *		void				SetTextWindowFont(
 *			WindowPtr			textWindowPtr,
 *			short				fontNumber
 *		);
 *		void				SetTextWindowFontSize(
 *			WindowPtr			textWindowPtr,
 *			short				fontSize
 *		);
 *		void				DisplayString(
 *			WindowPtr			textWindowPtr,
 *			StringPtr			theString
 *		);
 *		void				DisplayText(
 *			WindowPtr			textWindowPtr,
 *			Ptr					theText,
 *			short				byteOffset,
 *			short				byteCount
 *		);
 *
 * Description:
 *		This is a very simple logging window. You create
 *		one by calling NewTextWindow with the window
 *		title, the window shape and position, and a
 *		flag specifying whether the window has a
 *		goAway box. Text will be drawn in the application
 *		font (generally geneva) at 9 points. Edit the
 *		header file source to change this.
 *
 *		To write text into the window, call either
 *		DisplayString or DisplayText.
 *
 *		Add a call to TextWindowEvent to your event
 *		loop as follows:
 *			WaitNextEvent(...);
 *			textWindowState = TextWindowEvent(
 *						myTextWindow,
 *						&eventRecord
 *					);
 *			if ((textWindowState & kTextWindowEvent) == 0)
 *				Not a TextWindow event ...
 *			if ((textWindowState & kTextWindowMoved) != 0)
 *				Window moved, grew, or zoomed
 *		TextWindow handles grow, zoom, activate,
 *		suspend/resume, and update events.
 *		Your application handles other mouse events.
 *		Window content is not available to your
 *		application: you can't copy it to the clipboard,
 *		for example.
 *
 * GoAway box processing:
 *		If the user clicks in the goAway box, the library
 *		hides the window, but does not dispose of it.
 *		To dispose of the window, call DisposeTextWindow:
 *			if (((WindowPeek) textWindow)->visible
 *					== FALSE) {
 *				DisposeTextWindow(textWindow);
 *				textWindow = NULL;
 *			}
 *		To show the window, call
 *			ShowWindow((WindowPtr) myTextWindow);
 *
 * Errors:
 *		Memory errors are handled by the Failure library.
 */

#include <Types.h>
#include <Resources.h>
#include <QuickDraw.h>
#include <Fonts.h>
#include <Events.h>
#include <Windows.h>
#include <Menus.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Desk.h>
#include <ToolUtils.h>
#include <Memory.h>
#include <Files.h>
#include <OSUtils.h>
#include <OSEvents.h>
#include <DiskInit.h>
#include <Packages.h>
#include <Traps.h>
#include "SimpleTextWindow.h"
#include "Failure.h"
/*
 * Include DisplayAudit to get the error message list
 */
#include "DisplayAudit.h"

#define kTextWindowMinHeight	100
#define kTextWindowMinWidth		200
#define kTextWindowInset		  2
#define kScrollBarWidth			 16
#ifndef TRUE
#define TRUE					1
#define FALSE					0
#endif

#define height(box)				(box.bottom - box.top)
#define width(box)				(box.right - box.left)
		

typedef struct TextWindowRecord {
	WindowRecord		windowRecord;
	RgnHandle			updateRgn;
	Handle				*textHandlePtr;
	short				lineHeight;
	short				penOffset;
	short				fontAscent;
	short				nLines;
} TextWindowRecord, *TextWindowPtr;
/*
 * The current TextWindow will always be defined
 * by the textWindow local variable. Within code
 * sequences, the following globals may be used
 * to access variables:
 */
#define	TEXT_WINDOW		(*((TextWindowPtr) textWindowPtr))
#define TEXT_HANDLE		(TEXT_WINDOW.textHandlePtr)
#define PEN_OFFSET		(TEXT_WINDOW.penOffset)
#define FONT_ASCENT		(TEXT_WINDOW.fontAscent)
#define UPDATE_RGN		(TEXT_WINDOW.updateRgn)
#define LINE_HEIGHT		(TEXT_WINDOW.lineHeight)
#define NLINES			(TEXT_WINDOW.nLines)
#define PORT_RECT		(textWindowPtr->portRect)



static void					DrawMiniGrowIcon(
		register WindowPtr		textWindowPtr
	);
static void					MakeTextHandles(
		register WindowPtr		textWindowPtr
	);
static void					DisposeTextHandles(
		register WindowPtr		textWindowPtr
	);
static void					DrawTextWindow(
		register WindowPtr		textWindowPtr
	);
static void					StoreText(
		register WindowPtr		textWindowPtr,
		Ptr						theText,
		short					byteOffset,
		short					byteCount
	);
static void					ScrollText(
		register WindowPtr		textWindowPtr
	);

/*
 * Create a new text window.
 */
WindowPtr
NewTextWindow(
		StringPtr				windowTitle,
		Rect					*boundsRect,
		Boolean					hasGoAway,
		short					fontNumber,
		short					fontSize
	)
{
		WindowPtr			textWindowPtr;
		WindowPtr			theWindow;
		
		textWindowPtr = (WindowPtr) NewPtrClear(
				sizeof (TextWindowRecord)
			);
		FailNIL(textWindowPtr, kErrNoMemory);		
		theWindow = NewWindow(
				textWindowPtr,
				boundsRect,
				windowTitle,
				TRUE,
				zoomDocProc,
				(WindowPtr) -1L,
				hasGoAway,			/*Has go-away flag	*/
				0					/* refCon			*/
			);
		FailNIL(theWindow, kErrNoMemory);
		SelectWindow(theWindow);
		SetPort(theWindow);
		TextFont(fontNumber);
		TextSize(fontSize);
		UPDATE_RGN = NewRgn();
		FailNIL(UPDATE_RGN, kErrNoMemory);
		MakeTextHandles(textWindowPtr);
		return (textWindowPtr);
}

void
SetTextWindowFont(
		WindowPtr		textWindowPtr,
		short			fontNumber
	)
{
		SetPort(textWindowPtr);
		TextFont(fontNumber);
		MakeTextHandles(textWindowPtr);
}

void
SetTextWindowFontSize(
		WindowPtr		textWindowPtr,
		short			fontSize
	)
{
		SetPort(textWindowPtr);
		TextFont(fontSize);
		MakeTextHandles(textWindowPtr);
}


/*
 * Process all window events for the text window. Return
 * TRUE if we handled this event.
 *
 * Note, if the user clicks in the close box, we hide
 * the window and return TRUE. The caller should check
 * for a hidden window and call DisposeTextWindow to
 * actually release window storage:
 *		if (((WindowPeek) textWindow)->visible == FALSE) {
 *			DisposeTextWindow(textWindow);
 *			textWindow = NULL;
 *		}
 */
TextWindowState
TextWindowEvent(
		register WindowPtr		textWindowPtr,
		EventRecord				*theEvent
	)
{
		TextWindowState		textWindowState;		/* Our result	*/
		WindowPtr			eventWindow;
		short				windowPart;
		Rect				limitRect;
		GrafPtr				savePort;
		Boolean				isActivate;
		long				newSize;
		
		textWindowState = 0;
		switch (theEvent->what) {
		case mouseDown:
			windowPart = FindWindow(
							theEvent->where,
							&eventWindow
						);
			if (eventWindow == textWindowPtr) {
				switch (windowPart) {
				case inDrag:
					limitRect = qd.screenBits.bounds;
					DragWindow(
						textWindowPtr,
						theEvent->where,
						&limitRect
					);
					textWindowState |= (kTextWindowEvent | kTextWindowMoved);
					break;
				case inGrow:
					limitRect = qd.screenBits.bounds;
					limitRect.top = kTextWindowMinHeight;
					limitRect.left = kTextWindowMinWidth;
					newSize = GrowWindow(
								textWindowPtr,
								theEvent->where,
								&limitRect
							);
					if (newSize != 0) {
						InvalRect(&PORT_RECT);
						SizeWindow(
							textWindowPtr,
							LoWord(newSize),
							HiWord(newSize),
							TRUE
						);
						MakeTextHandles(textWindowPtr);
					}
					textWindowState |= (kTextWindowEvent | kTextWindowMoved);
					break;
				case inGoAway:
					HideWindow(textWindowPtr);
					textWindowState |= kTextWindowEvent;
					break;
				case inZoomIn:
				case inZoomOut:
					ZoomWindow(textWindowPtr, windowPart, TRUE);
					InvalRect(&PORT_RECT);
					MakeTextHandles(textWindowPtr);
					textWindowState |= (kTextWindowEvent | kTextWindowMoved);
					break;
				}
			}
			break;
		case updateEvt:
			if (((WindowPtr) theEvent->message) == textWindowPtr) {
				GetPort(&savePort);
				SetPort(textWindowPtr);
				BeginUpdate(textWindowPtr);
				EraseRect(&PORT_RECT);
				DrawMiniGrowIcon(textWindowPtr);
				DrawTextWindow(textWindowPtr);
				EndUpdate(textWindowPtr);
				SetPort(savePort);
				textWindowState |= kTextWindowEvent;
			}
			break;
		case activateEvt:
			if (((WindowPtr) theEvent->message)	== textWindowPtr) {
				textWindowState |= kTextWindowEvent;
				isActivate = theEvent->modifiers & activeFlag;
				goto doActivate;
			}
			break;
		case osEvt:
			/*
			 * Suspend/resume always returns FALSE as the
			 * application must activate/deactivate *all* of
			 * its windows on suspend/resume events. On the
			 * other hand, an activate event only pertains to
			 * the active window.
			 */
			if ((((unsigned long) theEvent->message) >> 24)
					== suspendResumeMessage) {
				isActivate =
					theEvent->message & resumeFlag;
doActivate:		if (isActivate && FrontWindow() == textWindowPtr) {
					ShowWindow(textWindowPtr);
					SelectWindow(textWindowPtr);
					SetPort(textWindowPtr);
				}
				DrawMiniGrowIcon(textWindowPtr);
			}
			break;
		}
		return (textWindowState);
}

void
DisplayString(
		register WindowPtr		textWindowPtr,
		StringPtr				theString
	)
{
		DisplayText(
			textWindowPtr,
			(Ptr) &theString[1],
			0,
			theString[0]
		);
}

void
DisplayText(
		register WindowPtr		textWindowPtr,
		Ptr						theText,
		short					byteOffset,
		short					byteCount
	)
{
		GrafPtr				savePort;
		
		GetPort(&savePort);
		SetPort(textWindowPtr);
		StoreText(
			textWindowPtr,
			theText,
			byteOffset,
			byteCount
		);
		ScrollText(textWindowPtr);
		MoveTo(kTextWindowInset, PEN_OFFSET);
		DrawText(theText, byteOffset, byteCount);
		SetPort(savePort);
}

static void
StoreText(
		register WindowPtr		textWindowPtr,
		Ptr						theText,
		short					byteOffset,
		short					byteCount
	)
{

		Handle					newLineHandle;
		int						i;
		
		/*
		 * Scroll all existing text up one line and
		 * move the top line handle to the bottom
		 * of the screen with its new contents.
		 */
		newLineHandle = TEXT_HANDLE[0];
		for (i = 1; i < NLINES; i++)
			TEXT_HANDLE[i - 1] = TEXT_HANDLE[i];
		FailOSErr(
			PtrToXHand(
				theText + byteOffset,
				newLineHandle,
				byteCount
			),
			kErrNoMemory
		);
		TEXT_HANDLE[NLINES - 1] = newLineHandle;
}

static void
ScrollText(
		register WindowPtr		textWindowPtr
	)
{

		Rect					activeRect;

		activeRect = PORT_RECT;
		activeRect.right -= kScrollBarWidth;
		ScrollRect(&activeRect, 0, -LINE_HEIGHT, UPDATE_RGN);
		EraseRgn(UPDATE_RGN);
		ValidRgn(UPDATE_RGN);
}

/*
 * Dispose of the window and its contents.
 */
void
DisposeTextWindow(
		register WindowPtr		textWindowPtr
	)
{
		RgnHandle				aRegionHdl;
		
		if (textWindowPtr != NULL) {
			CloseWindow(textWindowPtr);
			if ((aRegionHdl = UPDATE_RGN) != NULL) {
				UPDATE_RGN = NULL;
				DisposeRgn(aRegionHdl);
			}
			DisposeTextHandles(textWindowPtr);
			DisposPtr((Ptr) textWindowPtr);
		}
}

static void
MakeTextHandles(
		register WindowPtr		textWindowPtr
	)
{

		FontInfo			fontInfo;
		int					i;
		
		DisposeTextHandles(textWindowPtr);
		GetFontInfo(&fontInfo);
		LINE_HEIGHT =
					  fontInfo.ascent
					+ fontInfo.descent
					+ fontInfo.leading;
		NLINES = height(PORT_RECT) / LINE_HEIGHT;
		FONT_ASCENT = fontInfo.ascent;
		PEN_OFFSET =
			PORT_RECT.top
			+ ((NLINES - 1) * LINE_HEIGHT)
			+ fontInfo.ascent;
		/*
		 * TEXT_HANDLE is a vector of pointers
		 * to Handles. Each handle contains
		 * one line of text.
		 */
		TEXT_HANDLE = (Handle *)
			NewPtrClear(NLINES * sizeof (Handle));
		FailNIL(TEXT_HANDLE, kErrNoMemory);
		for (i = 0; i < NLINES; i++) {
			TEXT_HANDLE[i] = NewHandle(0);
			FailNIL(TEXT_HANDLE[i], kErrNoMemory);
		}
}

static void
DisposeTextHandles(
		register WindowPtr		textWindowPtr
	)
{
		int					i;
		Ptr					aPtr;
		Handle				aHandle;
		
		if ((aPtr = (Ptr) TEXT_HANDLE) != NULL) {
			for (i = 0; i < NLINES; i++) {
				aHandle = TEXT_HANDLE[i];
				if (aHandle != NULL) {
					TEXT_HANDLE[i] = NULL;
					DisposHandle(aHandle);
				}
			}
			TEXT_HANDLE = NULL;
			DisposePtr(aPtr);
		}		
}

static void
DrawTextWindow(
		register WindowPtr		textWindowPtr
	)
{
		int					i;
		short				offset;
		FontInfo			fontInfo;
		
		GetFontInfo(&fontInfo);
		offset = PORT_RECT.top + FONT_ASCENT;
		if (TEXT_HANDLE != NULL) {
			for (i = 0; i < NLINES; i++) {
				if (TEXT_HANDLE[i] != NULL) {
					MoveTo(kTextWindowInset, offset);
					HLock(TEXT_HANDLE[i]);
					DrawText(
						*TEXT_HANDLE[i],
						0,
						GetHandleSize(TEXT_HANDLE[i])
					);
					HUnlock(TEXT_HANDLE[i]);
				}
				offset += LINE_HEIGHT;
			}
		}
}

static void
DrawMiniGrowIcon(
		register WindowPtr		textWindowPtr
	)
{
		GrafPtr				savePort;
		PenState			savePen;
		RgnHandle			clipRgn;
		Rect				clipRect;

		GetPort(&savePort);
		SetPort(textWindowPtr);
		GetPenState(&savePen);
		PenNormal();
		clipRgn = NewRgn();
		if (clipRgn != NULL) {
			GetClip(clipRgn);
			clipRect = PORT_RECT;
			clipRect.left = clipRect.right - (kScrollBarWidth - 1);
			clipRect.top = clipRect.bottom - (kScrollBarWidth - 1);
			ClipRect(&clipRect);
			ValidRect(&clipRect);
			DrawGrowIcon(textWindowPtr);
			SetClip(clipRgn);
			DisposeRgn(clipRgn);
		}
		SetPenState(&savePen);
		SetPort(savePort);
}

